from __future__ import absolute_import

from uuid import uuid4

import numpy as np
import pandas as pd
import six
from bson import Code
from numpy import isscalar
from pandas.io.json import json_normalize
from pymongo.collection import Collection

from omegaml import defaults
from omegaml.store import qops
from omegaml.store.filtered import FilteredCollection
from omegaml.store.query import Filter, MongoQ
from omegaml.store.queryops import MongoQueryOps
from omegaml.util import make_tuple, make_list, restore_index, \
    cursor_to_dataframe, restore_index_columns_order, PickableCollection, extend_instance

INSPECT_CACHE = []


class MGrouper(object):
    """
    a Grouper for MDataFrames
    """
    STATS_MAP = {
        'std': 'stdDevSamp',
        'mean': 'avg',
    }

    def __init__(self, mdataframe, collection, columns, sort=True):
        self.mdataframe = mdataframe
        self.collection = collection
        self.columns = make_tuple(columns)
        self.should_sort = sort

    def __getattr__(self, attr):
        if attr in self.columns:
            return MSeriesGroupby(self, self.collection, attr)

        def statfunc():
            columns = self.columns or self._non_group_columns()
            return self.agg({col: attr for col in columns})

        return statfunc

    def __getitem__(self, item):
        return self.__getattr__(item)

    def agg(self, specs):
        """
        shortcut for .aggregate
        """
        return self.aggregate(specs)

    def aggregate(self, specs, **kwargs):
        """
        aggregate by given specs

        See the following link for a list of supported operations. 
        https://docs.mongodb.com/manual/reference/operator/aggregation/group/

        :param specs: a dictionary of { column : function | list[functions] } 
           pairs. 
        """

        def add_stats(specs, column, stat):
            specs['%s_%s' % (column, stat)] = {
                '$%s' % MGrouper.STATS_MAP.get(stat, stat): '$%s' % column}

        # generate $group command
        _specs = {}
        for column, stats in six.iteritems(specs):
            stats = make_tuple(stats)
            for stat in stats:
                add_stats(_specs, column, stat)
        groupby = qops.GROUP(columns=self.columns,
                             **_specs)
        # execute and return a dataframe
        pipeline = self._amend_pipeline([groupby])
        data = self.collection.aggregate(pipeline, allowDiskUse=True)

        def get_data():
            # we need this to build a pipeline for from_records
            # to process, otherwise the cursor will be exhausted already
            for group in data:
                _id = group.pop('_id')
                if isinstance(_id, dict):
                    group.update(_id)
                yield group

        df = pd.DataFrame.from_records(get_data())
        columns = make_list(self.columns)
        if columns:
            df = df.set_index(columns, drop=True)
        return df

    def _amend_pipeline(self, pipeline):
        """ amend pipeline with default ops on coll.aggregate() calls """
        if self.should_sort:
            sort = qops.SORT(**dict(qops.make_sortkey('_id')))
            pipeline.append(sort)
        return pipeline

    def _non_group_columns(self):
        """ get all columns in mdataframe that is not in columns """
        return [col for col in self.mdataframe.columns
                if col not in self.columns and col != '_id'
                and not col.startswith('_idx')
                and not col.startswith('_om#')]

    def _count(self):
        count_columns = self._non_group_columns()
        if len(count_columns) == 0:
            count_columns.append('_'.join(self.columns) + '_count')
        groupby = {
            "$group": {
                "_id": {k: "$%s" % k for k in self.columns},
            }
        }
        for k in count_columns:
            groupby['$group']['%s' % k] = {"$sum": 1}
        pipeline = self._amend_pipeline([groupby])
        if self.should_sort:
            sort = qops.SORT(**dict(qops.make_sortkey('_id')))
            pipeline.append(sort)
        return list(self.collection.aggregate(pipeline, allowDiskUse=True))

    def count(self):
        """ return counts by group columns """
        counts = self._count()
        # remove mongo object _id
        for group in counts:
            group.update(group.pop('_id'))
        # transform results to dataframe, then return as pandas would
        resultdf = pd.DataFrame(counts).set_index(make_list(self.columns),
                                                  drop=True)
        return resultdf

    def __iter__(self):
        """ for each group returns the key and a Filter object"""
        # reduce count to only one column
        groups = getattr(self, self.columns[0])._count()
        for group in groups:
            keys = group.get('_id')
            data = self.mdataframe._clone(query=keys)
            yield keys, data


class MLocIndexer(object):
    """
    implements the LocIndexer for MDataFrames
    """

    def __init__(self, mdataframe, positional=False):
        self.mdataframe = mdataframe
        # if positional, any loc[spec] will be applied on the rowid only
        self.positional = positional
        # indicator will be set true if loc specs are from a range type (list, tuple, np.ndarray)
        self._from_range = False

    def __getitem__(self, specs):
        """
        access by index

        use as mdf.loc[specs] where specs is any of

        * a list or tuple of scalar index values, e.g. .loc[(1,2,3)]
        * a slice of values e.g. .loc[1:5]
        * a list of slices, e.g. .loc[1:5, 2:3]

        :return: the sliced part of the MDataFrame
        """
        filterq, projection = self._get_filter(specs)
        df = self.mdataframe
        if filterq:
            df = self.mdataframe.query(filterq)
            df.from_loc_indexer = True
            df.from_loc_range = self._from_range
        if projection:
            df = df[projection]
        if isinstance(self.mdataframe, MSeries):
            df = df._as_mseries(df.columns[0])
        if getattr(df, 'immediate_loc', False):
            df = df.value
        return df

    def __setitem__(self, specs, value):
        raise NotImplemented

    def _get_filter(self, specs):
        filterq = []
        projection = []
        if self.positional:
            idx_cols = ['_om#rowid']
        else:
            idx_cols = self.mdataframe._get_frame_index()
        flt_kwargs = {}
        enumerable_types = (list, tuple, np.ndarray)
        if isinstance(specs, np.ndarray):
            specs = specs.tolist()
        if (isinstance(specs, enumerable_types)
                and isscalar(specs[0]) and len(idx_cols) == 1
                and not any(isinstance(s, slice) for s in specs)):
            # single column index with list of scalar values
            if (self.positional and isinstance(specs, tuple) and len(specs) == 2
                    and all(isscalar(v) for v in specs)):
                # iloc[int, int] is a cell access
                flt_kwargs[idx_cols[0]] = specs[0]
                projection.extend(self._get_projection(specs[1]))
            else:
                flt_kwargs['{}__in'.format(idx_cols[0])] = specs
                self._from_range = True
        elif isinstance(specs, (int, str)):
            flt_kwargs[idx_cols[0]] = specs
        else:
            specs = make_tuple(specs)
            # list/tuple of slices or scalar values, or MultiIndex
            for i, spec in enumerate(specs):
                if i < len(idx_cols):
                    col = idx_cols[i]
                    if isinstance(spec, slice):
                        self._from_range = True
                        start, stop = spec.start, spec.stop
                        if start is not None:
                            flt_kwargs['{}__gte'.format(col)] = start
                        if stop is not None:
                            if isinstance(stop, int):
                                stop -= int(self.positional)
                            flt_kwargs['{}__lte'.format(col)] = stop
                    elif isinstance(spec, enumerable_types) and isscalar(spec[0]):
                        self._from_range = True
                        # single column index with list of scalar values
                        # -- convert to list for PyMongo serialization
                        if isinstance(spec, np.ndarray):
                            spec = spec.tolist()
                        flt_kwargs['{}__in'.format(col)] = spec
                    elif isscalar(col):
                        flt_kwargs[col] = spec
                else:
                    # we're out of index columns, let's look at columns
                    projection.extend(self._get_projection(spec))
        if flt_kwargs:
            filterq.append(MongoQ(**flt_kwargs))
        finalq = None
        for q in filterq:
            if finalq:
                finalq |= q
            else:
                finalq = q
        return finalq, projection

    def _get_projection(self, spec):
        columns = self.mdataframe.columns
        if np.isscalar(spec):
            return [spec]
        if isinstance(spec, (tuple, list)):
            assert all(columns.index(col) for col in columns)
            return spec
        if isinstance(spec, slice):
            start, stop = spec.start, spec.stop
            if all(isinstance(v, int) for v in (start, stop)):
                start, stop, step = spec.indices(len(columns))
            else:
                start = columns.index(start) if start is not None else 0
                stop = columns.index(stop) + 1 if stop is not None else len(columns)
            return columns[slice(start, stop)]
        raise IndexError


class MPosIndexer(MLocIndexer):
    """
    implements the position-based indexer for MDataFrames
    """

    def __init__(self, mdataframe):
        super(MPosIndexer, self).__init__(mdataframe, positional=True)

    def _get_projection(self, spec):
        columns = self.mdataframe.columns
        if np.isscalar(spec):
            return columns[spec]
        if isinstance(spec, (tuple, list)):
            return [col for i, col in enumerate(spec) if i in spec]
        if isinstance(spec, slice):
            start, stop = slice.start, slice.stop
            if start and not isinstance(start, int):
                start = 0
            if stop and not isinstance(stop, int):
                # sliced ranges are inclusive
                stop = len(columns)
            return columns[slice(start, stop)]
        raise IndexError


class MSeriesGroupby(MGrouper):
    """
    like a MGrouper but limited to one column
    """

    def count(self):
        """
        return series count

        :return: counts by group 
        """
        # MGrouper will insert a _count column, see _count(). we remove
        # that column again and return a series named as the group column
        resultdf = super(MSeriesGroupby, self).count()
        count_column = [col for col in resultdf.columns
                        if col.endswith('_count')][0]
        new_column = count_column.replace('_count', '')
        resultdf = resultdf.rename(columns={count_column: new_column})
        return resultdf[new_column]


class MDataFrame(object):
    """
    A DataFrame for mongodb

    Performs out-of-core, lazy computOation on a mongodb cluster.
    Behaves like a pandas DataFrame. Actual results are returned
    as pandas DataFrames.
    """

    STATFUNCS = ['mean', 'std', 'min', 'max', 'sum', 'var']

    def __init__(self, collection, columns=None, query=None,
                 limit=None, skip=None, sort_order=None,
                 force_columns=None, immediate_loc=False, auto_inspect=False,
                 normalize=False, raw=False,
                 parser=None,
                 preparefn=None, **kwargs):
        self.collection = PickableCollection(collection)
        # columns in frame
        self.columns = make_tuple(columns) if columns else self._get_fields(raw=raw)
        self.columns = [str(col) for col in self.columns]
        # columns to sort by, defaults to not sorted
        self.sort_order = sort_order
        # top n documents to fetch
        self.head_limit = limit
        # top n documents to skip before returning
        self.skip_topn = skip
        # filter criteria
        self.filter_criteria = query or {}
        # force columns -- on output add columns not present
        self.force_columns = force_columns or []
        # was this created from the loc indexer?
        self.from_loc_indexer = kwargs.get('from_loc_indexer', False)
        # was the loc index used a range? Else a single value
        self.from_loc_range = None
        # setup query for filter criteries, if provided
        if self.filter_criteria:
            # make sure we have a filtered collection with the criteria given
            if isinstance(self.filter_criteria, dict):
                self.query_inplace(**self.filter_criteria)
            elif isinstance(self.filter_criteria, Filter):
                self.query_inplace(self.filter_criteria)
            else:
                raise ValueError('Invalid query specification of type {}'.format(type(self.filter_criteria)))
        # if immediate_loc is True, .loc and .iloc always evaluate
        self.immediate_loc = immediate_loc
        # __array__ will return this value if it is set, set it otherwise
        self._evaluated = None
        # set true to automatically capture inspects on .value. retrieve using .inspect(cached=True)
        self.auto_inspect = auto_inspect
        self._inspect_cache = INSPECT_CACHE
        # apply mixins
        self._applyto = str(self.__class__)
        self._apply_mixins()
        # parser to parse documents to dataframe
        self._parser = json_normalize if normalize else parser
        # prepare function to be applied just before returning from .value
        self._preparefn = preparefn
        # keep technical fields like _id, _idx etc
        self._raw = raw

    def _apply_mixins(self, *args, **kwargs):
        """
        apply mixins in defaults.OMEGA_MDF_MIXINS
        """
        from omegaml import settings
        defaults = settings()
        for mixin, applyto in defaults.OMEGA_MDF_MIXINS:
            if any(v in self._applyto for v in applyto.split(',')):
                extend_instance(self, mixin, *args, **kwargs)

    def __getstate__(self):
        # pickle support. note that the hard work is done in PickableCollection
        data = dict(self.__dict__)
        data.update(_evaluated=None)
        data.update(_inspect_cache=None)
        data.update(auto_inspect=self.auto_inspect)
        data.update(_preparefn=self._preparefn)
        data.update(_parser=self._parser)
        data.update(_raw=self._raw)
        data.update(collection=self.collection)
        return data

    def __setstate__(self, state):
        # pickle support. note that the hard work is done in PickableCollection
        self.__dict__.update(**state)

    def _getcopy_kwargs(self, without=None):
        """ return all parameters required on a copy of this MDataFrame """
        kwargs = dict(columns=self.columns,
                      sort_order=self.sort_order,
                      limit=self.head_limit,
                      skip=self.skip_topn,
                      from_loc_indexer=self.from_loc_indexer,
                      immediate_loc=self.immediate_loc,
                      query=self.filter_criteria,
                      auto_inspect=self.auto_inspect,
                      preparefn=self._preparefn)
        [kwargs.pop(k) for k in make_tuple(without or [])]
        return kwargs

    def __array__(self, dtype=None):
        # FIXME inefficient. make MDataFrame a drop-in replacement for any numpy ndarray
        # this evaluates every single time
        if self._evaluated is None:
            self._evaluated = array = self.value.as_matrix()
        else:
            array = self._evaluated
        return array

    def __getattr__(self, attr):
        if attr in MDataFrame.STATFUNCS:
            return self.statfunc(attr)
        if attr in self.columns:
            kwargs = self._getcopy_kwargs()
            kwargs.update(columns=attr)
            return MSeries(self.collection, **kwargs)
        raise AttributeError(attr)

    def __getitem__(self, cols_or_slice):
        """
        select and project by column, columns, slice, masked-style filter

        Masked-style filters work similar to pd.DataFrame/Series masks
        but do not actually return masks but an instance of Filter. A
        Filter is a delayed evaluation on the data frame.

            # select all rows where any column is == 5
            mdf = MDataFrame(coll)
            flt = mdf == 5
            mdf[flt]
            =>

        :param cols_or_slice: single column (str), multi-columns (list),
          slice to select columns or a masked-style
        :return: filtered MDataFrame or MSeries
        """
        if isinstance(cols_or_slice, six.string_types):
            # column name => MSeries
            return self._as_mseries(cols_or_slice)
        elif isinstance(cols_or_slice, int):
            # column number => MSeries
            column = self.columns[cols_or_slice]
            return self._as_mseries(column)
        elif isinstance(cols_or_slice, (tuple, list)):
            # list of column names => MDataFrame subset on columns
            kwargs = self._getcopy_kwargs()
            kwargs.update(columns=cols_or_slice)
            return MDataFrame(self.collection, **kwargs)
        elif isinstance(cols_or_slice, Filter):
            kwargs = self._getcopy_kwargs()
            kwargs.update(query=cols_or_slice.query)
            return MDataFrame(self.collection, **kwargs)
        elif isinstance(cols_or_slice, np.ndarray):
            raise NotImplemented
        raise ValueError('unknown accessor type %s' % type(cols_or_slice))

    def __setitem__(self, column, value):
        # True for any scalar type, numeric, bool, string
        if np.isscalar(value):
            result = self.collection.update_many(filter=self.filter_criteria,
                                                 update=qops.SET(column, value))
            self.columns.append(column)
        return self

    def _clone(self, collection=None, **kwargs):
        # convenience method to clone itself with updates
        return self.__class__(collection or self.collection, **kwargs,
                              **self._getcopy_kwargs(without=list(kwargs.keys())))

    def statfunc(self, stat):
        aggr = MGrouper(self, self.collection, [], sort=False)
        return getattr(aggr, stat)

    def groupby(self, columns, sort=True):
        """
        Group by a given set of columns

        :param columns: the list of columns
        :param sort: if True sort by group key
        :return: MGrouper
        """
        return MGrouper(self, self.collection, columns, sort=sort)

    def _get_fields(self, raw=False):
        result = []
        doc = self.collection.find_one()
        if doc is not None:
            if raw:
                result = list(doc.keys())
            else:
                result = [str(col) for col in doc.keys()
                          if col != '_id'
                          and not col.startswith('_idx')
                          and not col.startswith('_om#')]
        return result

    def _get_frame_index(self):
        """ return the dataframe's index columns """
        doc = self.collection.find_one()
        if doc is None:
            result = []
        else:
            result = restore_index_columns_order(doc.keys())
        return result

    def _get_frame_om_fields(self):
        """ return the dataframe's omega special fields columns """
        doc = self.collection.find_one()
        if doc is None:
            result = []
        else:
            result = [k for k in list(doc.keys()) if k.startswith('_om#')]
        return result

    def _as_mseries(self, column):
        kwargs = self._getcopy_kwargs()
        kwargs.update(columns=make_tuple(column))
        return MSeries(self.collection, **kwargs)

    def inspect(self, explain=False, cached=False, cursor=None, raw=False):
        """
        inspect this dataframe's actual mongodb query

        :param explain: if True explains access path
        """
        if not cached:
            if isinstance(self.collection, FilteredCollection):
                query = self.collection.query
            else:
                query = '*',
            if explain:
                cursor = cursor or self._get_cursor()
                explain = cursor.explain()
            data = {
                'projection': self.columns,
                'query': query,
                'explain': explain or 'specify explain=True'
            }
        else:
            data = self._inspect_cache
        if not (raw or explain):
            data = pd.DataFrame(pd.io.json.json_normalize(data))
        return data

    def count(self):
        """
        projected number of rows when resolving
        """
        counts = pd.Series({
            col: len(self)
            for col in self.columns}, index=self.columns)
        return counts

    def __len__(self):
        """
        the projected number of rows when resolving
        """
        return self._get_cursor().count()

    @property
    def shape(self):
        """
        return shape of dataframe
        """
        return len(self), len(self.columns)

    @property
    def ndim(self):
        return len(self.shape)

    @property
    def value(self):
        """
        resolve the query and return a Pandas DataFrame

        :return: the result of the query as a pandas DataFrame 
        """
        cursor = self._get_cursor()
        df = self._get_dataframe_from_cursor(cursor)
        if self.auto_inspect:
            self._inspect_cache.append(self.inspect(explain=True, cursor=cursor, raw=True))
        # this ensures the equiv. of pandas df.loc[n] is a Series
        if self.from_loc_indexer:
            if len(df) == 1 and not self.from_loc_range:
                idx = df.index
                df = df.T
                df = df[df.columns[0]]
                if df.ndim == 1 and len(df) == 1 and not isinstance(idx, pd.MultiIndex):
                    # single row single dimension, numeric index only
                    df = df.iloc[0]
            elif (df.ndim == 1 or df.shape[1] == 1) and not self.from_loc_range:
                df = df[df.columns[0]]
        if self._preparefn:
            df = self._preparefn(df)
        return df

    def reset(self):
        # TODO if head(), tail(), query(), .loc/iloc should return a new MDataFrame instance to avoid having a reset need
        self.head_limit = None
        self.skip_topn = None
        self.filter_criteria = {}
        self.force_columns = []
        self.sort_order = None
        self.from_loc_indexer = False
        return self

    def _get_dataframe_from_cursor(self, cursor):
        """ 
        from the given cursor return a DataFrame
        """
        df = cursor_to_dataframe(cursor, parser=self._parser)
        df = self._restore_dataframe_proper(df)
        return df

    def _restore_dataframe_proper(self, df):
        df = restore_index(df, dict())
        if '_id' in df.columns and not self._raw:
            df.drop('_id', axis=1, inplace=True)
        if self.force_columns:
            missing = set(self.force_columns) - set(self.columns)
            for col in missing:
                df[col] = np.NaN
        return df

    def _get_cursor(self):
        projection = make_tuple(self.columns)
        projection += make_tuple(self._get_frame_index())
        if not self.sort_order:
            # implicit sort
            projection += make_tuple(self._get_frame_om_fields())
        cursor = self.collection.find(projection=projection)
        if self.sort_order:
            cursor.sort(qops.make_sortkey(make_tuple(self.sort_order)))
        if self.head_limit:
            cursor.limit(self.head_limit)
        if self.skip_topn:
            cursor.skip(self.skip_topn)
        return cursor

    def sort(self, columns):
        """
        sort by specified columns

        :param columns: str of single column or a list of columns. Sort order
                        is specified as the + (ascending) or - (descending)
                        prefix to the column name. Default sort order is
                        ascending.
        :return: the MDataFrame
        """
        self._evaluated = None
        self.sort_order = make_tuple(columns)
        return self

    def head(self, limit=10):
        """
        return up to limit numbers of rows

        :param limit: the number of rows to return. Defaults to 10
        :return: the MDataFrame
        """
        return self._clone(limit=limit)

    def tail(self, limit=10):
        """
        return up to limit number of rows from last inserted values

        :param limit:
        :return:
        """
        tail_n = self.skip(len(self) - limit)
        return self._clone(skip=tail_n)

    def skip(self, topn):
        """
        skip the topn number of rows

        :param topn: the number of rows to skip.
        :return: the MDataFrame 
        """
        return self._clone(skip=topn)

    def merge(self, right, on=None, left_on=None, right_on=None,
              how='inner', target=None, suffixes=('_x', '_y'),
              sort=False, inspect=False):
        """
        merge this dataframe with another dataframe. only left outer joins
        are currently supported. the output is saved as a new collection,
        target name (defaults to a generated name if not specified).

        :param right: the other MDataFrame
        :param on: the list of key columns to merge by
        :param left_on: the list of the key columns to merge on this dataframe
        :param right_on: the list of the key columns to merge on the other 
           dataframe
        :param how: the method to merge. supported are left, inner, right. 
           Defaults to inner
        :param target: the name of the collection to store the merge results
           in. If not provided a temporary name will be created.
        :param suffixes: the suffixes to apply to identical left and right 
           columns
        :param sort: if True the merge results will be sorted. If False the
           MongoDB natural order is implied.
        :returns: the MDataFrame to the target MDataFrame
        """
        # validate input
        supported_how = ["left", 'inner', 'right']
        assert how in supported_how, "only %s merges are currently supported" % supported_how
        for key in [on, left_on, right_on]:
            if key:
                assert isinstance(
                    key, six.string_types), "only single column merge keys are supported (%s)" % key
        if isinstance(right, Collection):
            right = MDataFrame(right)
        assert isinstance(
            right, MDataFrame), "both must be MDataFrames, got right=%" % type(right)
        if how == 'right':
            # A right B == B left A
            return right.merge(self, on=on, left_on=right_on, right_on=left_on,
                               how='left', target=target, suffixes=suffixes)
        # generate lookup parameters
        on = on or '_id'
        right_name = self._get_collection_name_of(right, right)
        target_name = self._get_collection_name_of(
            target, '_temp.merge.%s' % uuid4().hex)
        target_field = (
                "%s_%s" % (right_name.replace('.', '_'), right_on or on))
        lookup = qops.LOOKUP(right_name,
                             key=on,
                             left_key=left_on,
                             right_key=right_on,
                             target=target_field)
        # unwind merged documents from arrays to top-level document fields
        unwind = qops.UNWIND(target_field, preserve=how != 'inner')
        # get all fields from left, right
        project = {}
        for left_col in self.columns:
            source_left_col = left_col
            if left_col == '_id':
                project[left_col] = 1
                continue
            if left_col.startswith('_idx'):
                continue
            if left_col.startswith('_om#'):
                continue
            if left_col != (on or left_on) and left_col in right.columns:
                left_col = '%s%s' % (left_col, suffixes[0])
            project[left_col] = "$%s" % source_left_col
        for right_col in right.columns:
            if right_col == '_id':
                continue
            if right_col.startswith('_idx'):
                continue
            if right_col.startswith('_om#'):
                continue
            if right_col == (on or right_on) and right_col == (on or left_on):
                # if the merge field is the same in both frames, we already
                # have it from left
                continue
            if right_col in self.columns:
                left_col = '%s%s' % (right_col, suffixes[1])
            else:
                left_col = '%s' % right_col
            project[left_col] = '$%s.%s' % (target_field, right_col)
        expected_columns = list(project.keys())
        project = {"$project": project}
        # store merged documents and return an MDataFrame to it
        out = qops.OUT(target_name)
        pipeline = [lookup, unwind, project]
        if sort:
            sort_cols = make_list(on or [left_on, right_on])
            sort_key = qops.make_sortkey(sort_cols)
            sort = qops.SORT(**dict(sort_key))
            pipeline.append(sort)
        pipeline.append(out)
        if inspect:
            result = pipeline
        else:
            result = self.collection.aggregate(pipeline, allowDiskUse=True)
            result = MDataFrame(self.collection.database[target_name],
                                force_columns=expected_columns)
        return result

    def append(self, other):
        if isinstance(other, Collection):
            right = MDataFrame(other)
        assert isinstance(
            other, MDataFrame), "both must be MDataFrames, got other={}".format(type(other))
        outname = self.collection.name
        mrout = {
            'merge': outname,
            'nonAtomic': True,
        }
        mapfn = Code("""
        function() {
           this._id = ObjectId();
           if(this['_om#rowid']) {
              this['_om#rowid'] += %s;
           }
           emit(this._id, this);
        }
        """ % len(self))
        reducefn = Code("""
        function(key, value) {
           return value;
        }
        """)
        finfn = Code("""
        function(key, value) {
           return value;
        }
        """)
        other.collection.map_reduce(mapfn, reducefn, mrout, finalize=finfn, jsMode=True)
        unwind = {
            "$replaceRoot": {
                "newRoot": {
                    "$ifNull": ["$value", "$$CURRENT"],
                }
            }
        }
        output = qops.OUT(outname)
        pipeline = [unwind, output]
        self.collection.aggregate(pipeline, allowDiskUse=True)
        return self

    def _get_collection_name_of(self, some, default=None):
        """
        determine the collection name of the given parameter

        returns the collection name if some is a MDataFrame, a Collection
        or a string_type. Otherwise returns default
        """
        if isinstance(some, MDataFrame):
            name = some.collection.name
        elif isinstance(some, Collection):
            name = some.name
        else:
            name = default
        return name

    def _get_filter_criteria(self, *args, **kwargs):
        """ 
        return mongo query from filter specs

        this uses a Filter to produce the query from the kwargs.

        :param args: a Q object or logical combination of Q objects
           (optional)
        :param kwargs: all AND filter criteria 
        """
        if len(args) > 0:
            q = args[0]
            if isinstance(q, MongoQ):
                filter_criteria = Filter(self.collection, q).query
            elif isinstance(q, Filter):
                filter_criteria = Filter(self.collection, q.q).query
        else:
            filter_criteria = Filter(self.collection, **kwargs).query
        return filter_criteria

    def query_inplace(self, *args, **kwargs):
        """
        filters this MDataFrame and returns it. 

        Any subsequent operation on the dataframe will have the filter
        applied. To reset the filter call .reset() without arguments.

        :param args: a Q object or logical combination of Q objects
           (optional)
        :param kwargs: all AND filter criteria 
        :return: self
        """
        self._evaluated = None
        self.filter_criteria = self._get_filter_criteria(*args, **kwargs)
        self.collection = FilteredCollection(
            self.collection, query=self.filter_criteria)
        return self

    def query(self, *args, **kwargs):
        """
        return a new MDataFrame with a filter criteria

        Any subsequent operation on the new dataframe will have the filter
        applied. To reset the filter call .reset() without arguments.

        Note: Unlike pandas DataFrames, a filtered MDataFrame operates
        on the same collection as the original DataFrame

        :param args: a Q object or logical combination of Q objects
           (optional)
        :param kwargs: all AND filter criteria 
        :return: a new MDataFrame with the filter applied
        """
        effective_filter = dict(self.filter_criteria)
        filter_criteria = self._get_filter_criteria(*args, **kwargs)
        if '$and' in effective_filter:
            effective_filter['$and'].extend(filter_criteria.get('$and'))
        else:
            effective_filter.update(filter_criteria)
        coll = FilteredCollection(self.collection, query=effective_filter)
        return self._clone(collection=coll, query=effective_filter)

    def create_index(self, keys, **kwargs):
        """
        create and index the easy way
        """
        keys, kwargs = MongoQueryOps().make_index(keys)
        result = self.collection.create_index(keys, **kwargs)
        return result

    def list_indexes(self):
        """
        list all indices in database
        """
        return cursor_to_dataframe(self.collection.list_indexes())

    @property
    def loc(self):
        """
        Access by index

        Use as mdf.loc[index_value]

        :return: MLocIndexer
        """
        self._evaluated = None
        indexer = MLocIndexer(self)
        return indexer

    @property
    def iloc(self):
        self._evaluated = None
        indexer = MPosIndexer(self)
        return indexer

    def __repr__(self):
        kwargs = ', '.join('{}={}'.format(k, v) for k, v in six.iteritems(self._getcopy_kwargs()))
        return "MDataFrame(collection={collection.name}, {kwargs})".format(collection=self.collection,
                                                                           kwargs=kwargs)


class MSeries(MDataFrame):
    """
    Series implementation for MDataFrames 

    behaves like a DataFrame but limited to one column.
    """

    def __init__(self, *args, **kwargs):
        super(MSeries, self).__init__(*args, **kwargs)
        # true if only unique values apply
        self.is_unique = False
        # apply mixins
        self._applyto = str(self.__class__)
        self._apply_mixins(*args, **kwargs)

    def __getitem__(self, cols_or_slice):
        if isinstance(cols_or_slice, Filter):
            return MSeries(self.collection, columns=self.columns,
                           query=cols_or_slice.query)
        return super(MSeries, self).__getitem__(cols_or_slice)

    @property
    def name(self):
        return self.columns[0]

    def unique(self):
        """
        return the unique set of values for the series

        :return: MSeries
        """
        self.is_unique = True
        return self

    def _get_cursor(self):
        if self.is_unique:
            # this way indexes get applied
            cursor = self.collection.distinct(make_tuple(self.columns)[0])
        else:
            cursor = super(MSeries, self)._get_cursor()
        return cursor

    @property
    def value(self):
        """
        return the value of the series

        this is a Series unless unique() was called. If unique()
        only distinct values are returned as an array, matching
        the behavior of a Series

        :return: pandas.Series 
        """
        cursor = self._get_cursor()
        column = make_tuple(self.columns)[0]
        if self.is_unique:
            # the .distinct() cursor returns a list of values
            # this is to make sure we return the same thing as pandas
            val = [v for v in cursor]
        else:
            val = self._get_dataframe_from_cursor(cursor)
            val = val[column]
            val.name = self.name
            if len(val) == 1 and self.from_loc_indexer:
                val = val.iloc[0]
        if self.auto_inspect:
            self._inspect_cache.append(self.inspect(explain=True, cursor=cursor, raw=True))
        if self._preparefn:
            df = self._preparefn(val)
        return val

    def __repr__(self):
        kwargs = ', '.join('{}={}'.format(k, v) for k, v in six.iteritems(self._getcopy_kwargs()))
        return "MSeries(collection={collection.name}, {kwargs})".format(collection=self.collection,
                                                                        kwargs=kwargs)



import os

import yaml
from selenium.webdriver.common.keys import Keys
from time import sleep
from urllib.parse import quote

istrue = lambda v: (
    (v.lower() in ('yes', '1', 'y', 'true', 't'))
    if isinstance(v, str) else bool(v)
)
isfalse = lambda v: not istrue(v)


def uri(browser, uri):
    """ given a browser, replace the path with uri """
    from six.moves.urllib.parse import urlparse, urlunparse
    url = browser.url
    parsed = list(urlparse(url))
    parsed[2] = uri
    return urlunparse(parsed)


ACTIVATE_CELL = Keys.ESCAPE, Keys.ENTER
EXEC_CELL = Keys.SHIFT, Keys.ENTER
ADD_CELL_BELOW = Keys.ESCAPE, 'b'
SAVE_NOTEBOOK = Keys.CONTROL, 's'


class Notebook:
    """
    A simple driver for the notebook
    """

    def __init__(self, browser):
        self.browser = browser
        try:
            alert = browser.get_alert()
        except:
            pass
        else:
            if alert is not None:
                alert.accept()

    @property
    def body(self):
        return self.browser.find_by_css('body').first

    @property
    def jupyter_home(self):
        br = self.browser
        br.windows.current = br.windows[0]
        sleep(1)
        return self

    @property
    def last_notebook(self):
        br = self.browser
        br.windows.current = br.windows[-1]
        return self

    def login_hub(self):
        br = self.browser
        br.find_by_id('username_input').first.fill('admin')
        br.find_by_id('password_input').first.fill('test')
        br.click_link_by_id('login_submit')
        assert br.is_element_present_by_id('ipython-main-app', wait_time=60)
        # check that there is actually a connection
        assert not br.is_text_present('Server error: Traceback', wait_time=5)
        assert not br.is_text_present('Connection refuse', wait_time=5)

    def login_nb(self):
        br = self.browser
        assert br.is_element_present_by_id('ipython-main-app', wait_time=2)
        br.find_by_id('password_input').fill('omegamlisfun')
        br.find_by_id('login_submit').click()
        # check that there is actually a connection
        assert not br.is_text_present('Server error: Traceback', wait_time=2)
        assert not br.is_text_present('Connection refuse', wait_time=2)

    def create_folder(self):
        """
        create a folder
        """
        br = self.browser
        self.jupyter_home
        br.find_by_id('new-dropdown-button').click()
        br.find_by_text('Folder').click()
        sleep(2)

    def create_notebook(self, folder=None):
        """
        create a new notebook
        """
        br = self.browser
        self.jupyter_home
        br.find_by_id('new-dropdown-button').click()
        br.find_by_text('Python 3').click()
        sleep(2)
        self.last_notebook
        return self

    def open_notebook(self, name, retry=5):
        self.jupyter_home
        br = self.browser
        retry = 5
        # FIXME sometimes it takes long for the nb to appear why?
        while retry:
            br.reload()
            found = br.is_text_present(name, wait_time=60)
            retry = 0 if found else retry
        item = br.find_link_by_partial_text(name)
        item.click()
        sleep(2)
        self.last_notebook
        return self

    def restart(self, wait=False):
        br = self.browser
        assert br.is_element_present_by_text('Cell', wait_time=5)
        br.find_link_by_text('Kernel', )[0].click()
        sleep(1)
        br.find_link_by_text('Restart')[0].click()
        if wait:
            busy = True
            while busy:
                sleep(5)
                busy = br.is_element_present_by_css('#kernel_indicator_icon.kernel_busy_icon')

    def run_all_cells(self, wait=False):
        br = self.browser
        assert br.is_element_present_by_text('Cell', wait_time=5)
        br.find_link_by_text('Cell', )[0].click()
        sleep(1)
        br.find_link_by_text('Run All')[0].click()
        if wait:
            busy = True
            while busy:
                sleep(5)
                busy = br.is_element_present_by_css('#kernel_indicator_icon.kernel_busy_icon')

    def open_folder(self, folder=None):
        br = self.browser
        folder = quote(folder.encode('utf-8'))
        item = br.find_link_by_href('/tree/{folder}'.format(**locals()))[0]
        item.click()
        return self

    def _clean_code(self, code):
        return tuple('\n'.join(line.strip() for line in code.split('\n')))

    def current_cell_exec(self, code):
        self.body.type(ACTIVATE_CELL + self._clean_code(code) + EXEC_CELL)

    def new_cell_exec(self, code):
        self.body.type(ADD_CELL_BELOW + ACTIVATE_CELL + self._clean_code(code) + EXEC_CELL)

    def current_cell_output(self):
        return self.body.find_by_css('.output_subarea pre')[-1].text

    def save_notebook(self):
        self.body.type(SAVE_NOTEBOOK)


def get_admin_secrets(scope=None, keys=None):
    secrets = os.path.join(os.path.expanduser('~/.omegaml/behave.yml'))
    with open(secrets) as fin:
        secrets = yaml.safe_load(fin)
        secrets = secrets[scope] if scope else secrets
    if keys:
        result = [secrets.get(k) for k in keys]
    else:
        result = secrets
    return result



import os

import nbformat
from datetime import datetime
from notebook.services.contents.manager import ContentsManager
from tornado import web
from urllib.parse import unquote

from omegaml.notebook.checkpoints import NoOpCheckpoints


class OmegaStoreContentsManager(ContentsManager):
    """
    Jupyter notebook storage manager for omegaml

    This requires a properly configured omegaml instance.

    see http://jupyter-notebook.readthedocs.io/en/stable/extending/contents.html
    """

    def __init__(self, **kwargs):
        # pass omega= for testing purpose
        self._omega = kwargs.pop('omega', None)
        super(OmegaStoreContentsManager, self).__init__(**kwargs)

    def _checkpoints_class_default(self):
        return NoOpCheckpoints

    @property
    def omega(self):
        """
        return the omega instance used by the contents manager
        """
        if self._omega is None:
            import omegaml as om
            self._omega = om
        self._omega.jobs._include_dir_placeholder = True
        return self._omega

    @property
    def store(self):
        """
        return the OmageStore for jobs (notebooks)
        """
        return self.omega.jobs.store

    @property
    def _dir_placeholder(self):
        return self.omega.jobs._dir_placeholder

    def get(self, path, content=True, type=None, format=None):
        """
        get an entry in the store

        this is called by the contents engine to get the contents of the jobs
        store.
        """
        path = unquote(path).strip('/')
        if not self.exists(path):
            raise web.HTTPError(404, u'No such file or directorys: %s' % path)

        if path == '' or type == 'directory':
            model = self._dir_model(path, content=content)
        elif type == 'notebook' or (type is None and path.endswith('.ipynb')):
            model = self._notebook_model(path, content=content)
        else:
            raise web.HTTPError(404, u'Type {} at {} is not supported'.format(type, path))
        return model

    def save(self, model, path):
        """
        save an entry in the store

        this is called by the contents engine to store a notebook
        """
        om = self.omega
        path = unquote(path).strip('/')
        type = model.get('type')
        name = model.get('name')
        self.run_pre_save_hook(model=model, path=path)
        if type is None:
            raise web.HTTPError(400, u'No file type provided')
        try:
            if type == 'notebook':
                content = model.get('content')
                if content is None:
                    raise web.HTTPError(400, u'No file content provided')
                nb = nbformat.from_dict(model['content'])
                self.check_and_sign(nb, path)
                self.omega.jobs.put(nb, path)
                self.validate_notebook_model(model)
                validation_message = model.get('message', None)
                model = self.get(path, content=False, type=type)
                if validation_message:
                    model['message'] = validation_message
            elif type == 'directory':
                ph_name = '{path}/{self._dir_placeholder}'.format(**locals()).strip('/')
                self.omega.jobs.create("#placeholder", ph_name)
                model = self.get(path, content=False, type=type)
                model['content'] = None
                model['format'] = None
                validation_message = None
            else:
                raise web.HTTPError(
                    400, "Unhandled contents type: %s" % model['type'])
        except web.HTTPError:
            raise
        except Exception as e:
            self.log.error(
                u'Error while saving file: %s %s', path, e, exc_info=True)
            raise web.HTTPError(
                500, u'Unexpected error while saving file: %s %s' % (path, e))
        return model

    def delete_file(self, path):
        """
        delete an entry

        this is called by the contents engine to delete an entry
        """
        path = unquote(path).strip('/')
        try:
            self.omega.jobs.drop(path)
        except Exception as e:
            self.omega.jobs.drop(path + '/' + self._dir_placeholder)

    def rename_file(self, old_path, new_path):
        """
        rename a file

        this is called by the contents engine to rename an entry
        """
        old_path = unquote(old_path).strip('/')
        new_path = unquote(new_path).strip('/')
        # check file or directory
        if self.file_exists(new_path):
            raise web.HTTPError(409, u'Notebook already exists: %s' % new_path)
        elif self.dir_exists(new_path):
            raise web.HTTPError(409, u'Directory already exists: %s' % new_path)
        # do the renaming
        dirname = old_path + '/' + self._dir_placeholder
        if self.dir_exists(dirname):
            meta = self.omega.jobs.metadata(dirname)
            meta.name = new_path + '/' + self._dir_placeholder
        elif self.file_exists(old_path):
            meta = self.omega.jobs.metadata(old_path)
            meta.name = new_path
        # rename on metadata. Note the gridfile instance stays the same
        meta.save()

    def exists(self, path):
        """
        Does a file or dir exist at the given collection in gridFS?
        We do not have dir so dir_exists returns true.

        :param path: (str) The relative path to the file's directory
          (with '/' as separator)
        :returns exists: (boo) The relative path to the file's directory (with '/' as separator)
        """
        path = unquote(path).strip('/')
        return self.file_exists(path) or (self.dir_exists(path))

    def dir_exists(self, path=''):
        """check if directory exists

        Args:
            path: name of directory

        Returns:
            True if directory exists
        """
        path = unquote(path).strip('/')
        if path == '':
            return True
        pattern = r'^{path}.*'.format(path=path)
        return len(self.omega.jobs.list(pattern)) > 0

    def file_exists(self, path):
        """check if file exists

        Args:
            path: name of file

        Returns:
            True if file exists
        """
        path = unquote(path).strip('/')
        return path in self.omega.jobs.list(path)

    def is_hidden(self, path):
        """check if path or file is hidden

        Args:
            path: name of file or path

        Returns:
            False, currently always returns false
        """
        return False

    def _read_notebook(self, path, as_version=None):
        path = unquote(path).strip('/')
        return self.omega.jobs.get(path)

    def _notebook_model(self, path, content=True):
        """
        Build a notebook model
        if content is requested, the notebook content will be populated
        as a JSON structure (not double-serialized)
        """
        path = unquote(path).strip('/')
        model = self._base_model(path)
        model['type'] = 'notebook'
        if content:
            nb = self._read_notebook(path, as_version=4)
            self.mark_trusted_cells(nb, path)
            model['content'] = nb
            model['format'] = 'json'
            self.validate_notebook_model(model)
        # always add accurate created and modified
        meta = self.omega.jobs.metadata(path)
        if meta is not None:
            model['created'] = meta.created
            model['last_modified'] = meta.modified
        else:
            model['last_modified'] = datetime.utcnow()
            model['created'] = datetime.utcnow()
        return model

    def _base_model(self, path, kind=None):
        """Build the common base of a contents model"""
        # http://jupyter-notebook.readthedocs.io/en/stable/extending/contents.html
        path = unquote(path).strip('/')
        last_modified = datetime.utcnow()
        created = last_modified
        # Create the base model.
        model = {}
        model['name'] = os.path.basename(path)
        model['path'] = path
        model['last_modified'] = last_modified
        model['created'] = created
        model['content'] = None
        model['format'] = None
        model['mimetype'] = None
        model['writable'] = True
        if kind:
            model['type'] = kind
            model['content'] = [] if kind == 'directory' else None
        return model

    def _dir_model(self, path, content=True):
        """
        Build a model to return all of the files in gridfs
        if content is requested, will include a listing of the directory
        """
        # this looks like a seemingly simple task, it's carefully crafted
        path = unquote(path).strip('/')
        model = self._base_model(path, kind='directory')
        model['format'] = 'json'
        contents = model['content']
        # get existing entries from a pattern that matches either
        #    top-level files: ([\w -]+\.[\w -]*)
        #    directories (files in): ([\w ]+/([\w ]+\.[\w]*))
        # does not work:
        #    pattern = r'([\w ]+/_placeholder\.[\w]*)|([\w ]+\.[\w]*)$'
        #    it is too restrictive as entries can be generated without a placeholder
        # so we get all, which may include sub/sub/file
        # and we need to include them because we need to find sub/sub directories
        # note \w is any word character (letter, digit, underscore)
        #      \s is any white space
        #      -  is the dash, literally
        pattern = r'([\w\s-]+\/)?([\w\s-]+\.[\w]*)$'
        # if we're looking in an existing directory, prepend that
        if path:
            pattern = '{path}/{pattern}'.format(path=path, pattern=pattern)
        pattern = '^{}'.format(pattern)
        entries = self.omega.jobs.list(pattern, raw=True)
        # by default assume the current path is listed already
        directories = [path]
        for meta in entries:
            # get path of entry, e.g. sub/foo.ipynb => sub
            entry_path = os.path.dirname(meta.name)
            # if not part of listed directories yet, include
            if entry_path not in directories:
                entry = self._base_model(entry_path, kind='directory')
                contents.append(entry)
                directories.append(entry_path)
            # ignore placeholder files
            if meta.name.endswith(self._dir_placeholder):
                continue
            # only include files that are in the path we're listing
            if entry_path != path:
                continue
            # include the actual file
            try:
                entry = self.get(meta.name, content=content, type='notebook')
            except Exception as e:
                msg = ('_dir_model error, cannot get {}, '
                       'removing from list, exception {}'.format(meta.name, str(e)))
                self.log.warning(msg)
            else:
                contents.append(entry)
        return model



import os
import six
from mongoengine import GridFSProxy

from omegaml.backends.basedata import BaseDataBackend


class PythonRawFileBackend(BaseDataBackend):
    """
    OmegaStore backend to support arbitrary files
    """
    KIND = 'python.file'

    @classmethod
    def supports(self, obj, name, as_raw=None, **kwargs):
        is_filelike = hasattr(obj, 'read')
        return is_filelike or self._is_path(self, obj)

    def _is_path(self, obj):
        return isinstance(obj, six.string_types) and os.path.exists(obj)

    def get(self, name, version=-1, lazy=False, **kwargs):
        return self.data_store.metadata(name, **kwargs).gridfile

    def put(self, obj, name, attributes=None, encoding=None, **kwargs):
        self.data_store.drop(name, force=True)
        fn = self.data_store._get_obj_store_key(name, 'file')
        if self._is_path(obj):
            with open(obj, 'rb') as fin:
                fileid = self.data_store.fs.put(fin, filename=fn)
        else:
            fileid = self.data_store.fs.put(obj, filename=fn, encoding=encoding)
        gridfile = GridFSProxy(grid_id=fileid,
                               db_alias='omega',
                               collection_name=self.data_store.bucket)
        return self.data_store._make_metadata(
            name=name,
            prefix=self.data_store.prefix,
            bucket=self.data_store.bucket,
            kind=self.KIND,
            attributes=attributes,
            gridfile=gridfile).save()



from pymongo import MongoClient as RealMongoClient


def MongoClient(*args, **kwargs):
    """
    Shim function adding SSL kwargs on MongoClient instead of changing 
    each and every location
    """
    from omegaml import settings
    defaults = settings()
    kwargs.update(defaults.OMEGA_MONGO_SSL_KWARGS)
    return RealMongoClient(*args, **kwargs)


from behave import given, when, then

from omegaml.tests.features.util import get_admin_secrets


@given('we have a connection to omegaml')
def connection(ctx):
    import omegaml as om
    ctx.feature.om = om.setup()
    assert ctx.feature.om is not None


@when('we ingest data')
def ingest(ctx):
    import pandas as pd
    om = ctx.feature.om
    data = {
        'x': range(10),
    }
    df = pd.DataFrame(data)
    df['y'] = df['x'] * 2
    om.datasets.put(df, 'sample', append=False)
    assert 'sample' in om.datasets.list()


@when('we build a model')
def model(ctx):
    from sklearn.linear_model import LinearRegression
    om = ctx.feature.om
    reg = LinearRegression()
    om.models.put(reg, 'regmodel')
    resp = om.runtime.model('regmodel').fit('sample[x]', 'sample[y]')
    resp.get()


@then('we can predict a result')
def predict(ctx):
    import numpy as np
    om = ctx.feature.om
    resp = om.runtime.model('regmodel').predict([1])
    result = resp.get()
    assert np.isclose(result[0], 2), "expected approx. 2, got {}".format(result[0])
    resp = om.runtime.model('regmodel').predict([50])
    result = resp.get()
    assert np.isclose(result[0], 100), "expected approx. 100, got {}".format(result[0])


@when('we store {scope} credentials in {dataset}')
def store_secrets(ctx, scope, dataset):
    om = ctx.feature.om
    secrets = get_admin_secrets(scope)
    om.datasets.put(secrets, dataset, append=False)
    assert 'secrets' in om.datasets.list()



import six
from flask_restplus import Model


class StrictModel(Model):
    # To implement a model that supports strict validation
    # on fields, we need to explicitly add into the schema
    # 'additionalProperties: False'
    # See: https://github.com/noirbizarre/flask-restplus/issues/241
    @property
    def _schema(self):
        old = super(StrictModel, self)._schema
        old['additionalProperties'] = False
        return old


class strict(object):
    # a poor man's stand-in for api.model
    def __init__(self, api):
        self.api = api

    def model(self, name=None, model=None, mask=None, **kwargs):
        # create a strict model and attach to api as in flask-restplus issue #241
        smodel = StrictModel(name, model, mask=mask)
        smodel.__apidoc__.update(kwargs)
        self.api.models[smodel.name] = smodel
        return smodel




import numpy as np
from six import BytesIO

from omegaml.backends.basedata import BaseDataBackend


class NumpyNDArrayBackend(BaseDataBackend):
    """
    Store numpy NDArray of any shape or size

    The NDArray is serialized to a byte string and stored as a BLOB.
    Thus it can have arbitrary size and dimensions, ideal for image
    data and Tensors.
    """
    KIND = 'ndarray.bin'

    _save_method = 'np.save'

    @classmethod
    def supports(self, obj, name, **kwargs):
        return isinstance(obj, np.ndarray)

    def put(self, obj, name, attributes=None, allow_pickle=False, **kwargs):
        # TODO associate meta.gridfile with actual fs file
        kind_meta = {
            'dtype': obj.dtype.name,
            'shape': obj.shape,
            'save_method': self._save_method,
            'allow_pickle': allow_pickle,
        }
        fout = self.data_store.fs.new_file(filename=name)
        buf = BytesIO()
        np.save(buf, obj, allow_pickle=allow_pickle)
        buf.seek(0)
        fout.write(buf)
        fout.close()
        return self.data_store.make_metadata(name, self.KIND, attributes=attributes,
                                             kind_meta=kind_meta,
                                             **kwargs).save()

    def get(self, name, version=-1, force_python=False, lazy=False, **kwargs):
        # TODO read file from meta.gridfile if exists, fallback to .fs only if not
        fin = self.data_store.fs.get_last_version(name)
        meta = self.data_store.metadata(name)
        if meta.kind_meta.get('save_method') == self._save_method:
            allow_pickle = meta.kind_meta.get('allow_pickle')
            loaded = self._load_from_npsave(fin, allow_pickle)
        else:
            dtype_name = meta.kind_meta['dtype']
            dtype = getattr(np, dtype_name, None)
            shape = meta.kind_meta['shape']
            loaded = self._load_from_stringbuffer(fin, dtype, shape)
        return loaded

    def _load_from_stringbuffer(self, fin, dtype, shape):
        return np.frombuffer(fin.read(), dtype=dtype).reshape(*shape)

    def _load_from_npsave(self, fin, allow_pickle):
        buf = BytesIO()
        buf.write(fin.read())
        fin.close()
        buf.seek(0)
        return np.load(buf, allow_pickle=allow_pickle)



from unittest import TestCase, skip

from omegaml import Omega
from omegaml.tests.cli.scenarios import CliTestScenarios
from omegaml.tests.util import OmegaTestMixin


class CliModelsTest(CliTestScenarios, OmegaTestMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.om = Omega()
        self.clean()

    @skip("only works in enterprise context")
    def test_cli_cloud_add_resource(self):
        self.cli('--config ../../../config.yml cloud add worker')



from unittest import TestCase

from omegaml.client.auth import OmegaRestApiAuth


class OmegaRestApiAuthTests(TestCase):
    def test_omegaauth(self):
        """ test OmegaRestApiAuth works ok
        """
        auth = OmegaRestApiAuth("foo", "bar")
        self.assertEqual(auth.username, 'foo')
        self.assertEqual(auth.apikey, 'bar')
        self.assertEqual(repr(auth), 'OmegaRestApiAuth(username=foo, apikey="*****",qualifier=default)')

    def test_omegaauth_qualifier(self):
        auth = OmegaRestApiAuth("foo", "bar", qualifier='xyz')
        self.assertEqual(auth.username, 'foo')
        self.assertEqual(auth.apikey, 'bar')
        self.assertEqual(repr(auth), 'OmegaRestApiAuth(username=foo, apikey="*****",qualifier=xyz)')

    def test_omegaauth_headers(self):
        class Request:
            headers = {}

        auth = OmegaRestApiAuth("foo", "bar", qualifier='xyz')
        self.assertEqual(auth.get_credentials(), 'ApiKey foo:bar')
        r = Request()
        auth(r)
        self.assertEqual(r.headers['Authorization'], 'ApiKey foo:bar')
        self.assertEqual(r.headers['Qualifier'], 'xyz')




from unittest import TestCase

import numpy as np

from omegaml import Omega
from omegaml.backends.tensorflow import TensorflowSavedModelPredictor
from omegaml.backends.tensorflow.tfkeras import TensorflowKerasBackend
from omegaml.backends.tensorflow.tfkerassavedmodel import TensorflowKerasSavedModelBackend
from omegaml.tests.util import OmegaTestMixin, tf_perhaps_eager_execution


class TensorflowKerasBackendTests(OmegaTestMixin, TestCase):
    def setUp(self):
        self.om = Omega()
        self.om.models.register_backend(TensorflowKerasBackend.KIND, TensorflowKerasBackend)
        self.om.models.register_backend(TensorflowKerasSavedModelBackend.KIND, TensorflowKerasSavedModelBackend)
        self.clean()
        tf_perhaps_eager_execution()

    def _build_model(self, fit=False):
        # build a dummy model for testing. does not need to make sense
        import tensorflow as tf
        keras = tf.keras
        Sequential = keras.models.Sequential
        Dense = keras.layers.Dense
        Dropout = keras.layers.Dropout
        SGD = keras.optimizers.SGD

        # Generate dummy data
        import numpy as np
        x_train = np.random.random((1000, 20))
        y_train = keras.utils.to_categorical(np.random.randint(10, size=(1000, 1)), num_classes=10)
        x_test = np.random.random((100, 20))
        y_test = keras.utils.to_categorical(np.random.randint(10, size=(100, 1)), num_classes=10)

        model = Sequential()
        # Dense(64) is a fully-connected layer with 64 hidden units.
        # in the first layer, you must specify the expected input data shape:
        # here, 20-dimensional vectors.
        model.add(Dense(64, activation='relu', input_dim=20))
        model.add(Dropout(0.5))
        model.add(Dense(64, activation='relu'))
        model.add(Dropout(0.5))
        model.add(Dense(10, activation='softmax', name='output'))
        sgd = SGD(lr=0.01, decay=1e-6, momentum=0.9, nesterov=True)
        model.compile(loss='categorical_crossentropy',
                      optimizer=sgd,
                      metrics=['accuracy'])

        if fit:
            model.fit(x_train, y_train,
                      epochs=1,
                      batch_size=128)
        return model

    def test_save_load(self):
        om = self.om
        model = self._build_model(fit=True)
        om.models.put(model, 'keras-model')
        model_ = om.models.get('keras-model')
        self.assertTrue(np.all(np.equal(model_.get_weights()[0], (model.get_weights()[0]))))

    def test_fit(self):
        from tensorflow import keras
        om = self.om
        model = self._build_model(fit=False)
        om.models.put(model, 'keras-model')
        x_test = np.random.random((100, 20))
        y_test = keras.utils.to_categorical(np.random.randint(10, size=(100, 1)), num_classes=10)
        result = om.runtime.model('keras-model').fit(x_test, y_test).get()
        self.assertTrue(result.startswith('<Metadata:'))
        result = om.runtime.model('keras-model').predict(x_test, epochs=10).get()
        self.assertEqual(result.shape, (100, 10))

    def test_fit_tpu(self):
        from tensorflow import keras
        om = self.om
        model = self._build_model(fit=False)
        om.models.put(model, 'keras-model')
        x_test = np.random.random((100, 20))
        y_test = keras.utils.to_categorical(np.random.randint(10, size=(100, 1)), num_classes=10)
        result = om.runtime.model('keras-model').fit(x_test, y_test, tpu_specs=True).get()
        self.assertTrue(result.startswith('<Metadata:'))
        result = om.runtime.model('keras-model').predict(x_test).get()
        self.assertEqual(result.shape, (100, 10))

    def test_runtime_predict_from_trained_model(self):
        om = self.om
        model = self._build_model(fit=True)
        om.models.put(model, 'keras-model')
        x_test = np.random.random((100, 20))
        result = om.runtime.model('keras-model').predict(x_test).get()
        self.assertEqual(result.shape, (100, 10))

    def test_save_load_savedmodel(self):
        om = self.om
        model = self._build_model(fit=True)
        x_test = np.random.random((100, 20))
        yhat = model.predict(x_test)
        om.models.put(model, 'keras-savedmodel', as_savedmodel=True)
        model_ = om.models.get('keras-savedmodel')
        self.assertIsInstance(model_, TensorflowSavedModelPredictor)
        yhat_ = model_.predict(x_test)
        self.assertTrue(np.allclose(yhat_, yhat))









from unittest import TestCase

from numpy.testing import assert_array_equal
from pandas.util.testing import assert_frame_equal

from omegaml import Omega
from omegaml.tests.cli.scenarios import CliTestScenarios
from omegaml.tests.util import OmegaTestMixin
from omegaml.util import temp_filename

import pandas as pd

class CliModelsTest(CliTestScenarios, OmegaTestMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.om = Omega()
        self.clean()

    def test_cli_datasets_help(self):
        self.cli('help datasets')

    def test_cli_datasets_list(self):
        # start with empty datasets
        self.cli('datasets list')
        expected = self.pretend_log([])
        self.assertLogEqual('info', expected)
        # add a test dataset
        self.make_dataset_from_dataframe('testds')
        self.cli('datasets list', new_start=True)
        expected = self.pretend_log(['testds'])
        self.assertLogEqual('info', expected)
        # add another one, use pattern matching
        self.make_dataset_from_dataframe('foobar')
        # -- just on one
        self.cli('datasets list test*', new_start=True)
        expected = self.pretend_log(['testds'])
        self.assertLogEqual('info', expected)
        # -- on the other
        self.cli('datasets list foo*', new_start=True)
        expected = self.pretend_log(['foobar'])
        self.assertLogEqual('info', expected)
        # match both
        self.cli('datasets list foo.*|test.* -E', new_start=True)
        expected = self.pretend_log(['foobar', 'testds'])
        self.assertLogEqual('info', expected)

    def test_cli_datasets_put(self):
        # write a csv, read back
        fn = temp_filename(ext='csv')
        df = self.create_local_csv(fn)
        self.cli(f'datasets put {fn} foobar')
        self.assertLogContains('info', 'Metadata(')
        self.assertLogContains('info', 'kind=pandas.dfrows')
        dfx = self.om.datasets.get('foobar')
        assert_frame_equal(df, dfx)
        # write a csv, read back, use custom sep
        fn = temp_filename(ext='csv')
        df = self.create_local_csv(fn, sep=';')
        self.cli(f'datasets put {fn} foxbaz --csv sep=;')
        self.assertLogContains('info', 'Metadata(')
        self.assertLogContains('info', 'kind=pandas.dfrows')
        dfx = self.om.datasets.get('foxbaz')
        assert_frame_equal(df, dfx)
        # write some other file
        fn = temp_filename(ext='binary')
        df = self.create_local_csv(fn)
        self.cli(f'datasets put {fn} binary', new_start=True)
        print(self.get_log('info'))
        self.assertLogContains('info', 'Metadata(')
        self.assertLogContains('info', 'kind=python.file')
        with open(fn, 'rb') as fin:
            expected = fin.read()
        data = self.om.datasets.get('binary').read()
        self.assertEqual(expected, data)
        # write image file
        fn = temp_filename(ext='jpg')
        img = self.create_local_image_file(fn)
        self.cli(f'datasets put {fn} image', new_start=True)
        self.assertLogContains('info', 'Metadata(')
        self.assertLogContains('info', 'kind=ndarray.bin')
        data = self.om.datasets.get('image')
        assert_array_equal(data, img)

    def test_cli_datasets_get(self):
        # write a csv, read back
        fn = temp_filename(ext='csv')
        self.make_dataset_from_dataframe('test')
        self.cli(f'datasets get test {fn}')
        df = pd.read_csv(fn)
        expected = self.om.datasets.get('test')
        assert_frame_equal(df, expected)
        # write a csv, read back with options
        fn = temp_filename(ext='csv')
        self.make_dataset_from_dataframe('test-sep')
        self.cli(f'datasets get test-sep {fn} --csv sep=; --csv columns=x,')
        df = pd.read_csv(fn, sep=';')
        expected = self.om.datasets.get('test')[['x']]
        assert_frame_equal(expected, df)

    def test_cli_datasets_drop(self):
        self.make_dataset_from_dataframe('test')
        self.assertIn('test', self.om.datasets.list())
        self.cli('datasets drop test')
        self.assertNotIn('test', self.om.datasets.list())

    def test_cli_datasets_metadata(self):
        self.make_dataset_from_dataframe('test')
        self.cli('datasets metadata test')
        expected = self.pretend_log('"kind": "pandas.dfrows"')
        self.assertLogContains('info', expected)
        expected = self.pretend_log('"name": "test"')
        self.assertLogContains('info', expected)















import pathlib
from unittest import TestCase, skipIf, skip

from omegaml import Omega
from omegaml.backends.tensorflow.tfdataset import TFDatasetBackend
from omegaml.tests.util import tf_in_eager_execution, tf_perhaps_eager_execution, OmegaTestMixin

# check https://www.tensorflow.org/datasets/api_docs/python/tfds/testing/run_in_graph_and_eager_modes
@skip("requires eager mode which must be enabled once for the whole python session")
class TensorflowDatasetBackendTests(OmegaTestMixin, TestCase):
    def setUp(self):
        import os
        os.environ['TF_EAGER'] = '1'
        tf_perhaps_eager_execution()
        self.om = Omega()
        self.om.models.register_backend(TFDatasetBackend.KIND, TFDatasetBackend)
        self.clean()

    def get_image_ds(self):
        import tensorflow as tf
        from tensorflow.python.data.experimental import AUTOTUNE

        data_root_orig = tf.keras.utils.get_file('flower_photos',
                                                 'https://storage.googleapis.com/download.tensorflow.org/example_images/flower_photos.tgz',
                                                 untar=True)
        data_root = pathlib.Path(data_root_orig)

        all_image_paths = list(data_root.glob('*/*'))
        all_image_paths = [str(path) for path in all_image_paths]

        def load_and_preprocess_image(path):
            image = tf.read_file(path)
            return image

        path_ds = tf.data.Dataset.from_tensor_slices(all_image_paths)
        image_ds = path_ds.map(load_and_preprocess_image, num_parallel_calls=AUTOTUNE)
        return image_ds

    def test_save_dataset(self):
        import tensorflow as tf
        ds = self.get_image_ds()
        for img in ds.take(1):
            print(img.numpy())
            break








import logging
import unittest

from omegaml import Omega
from omegaml.store.logging import OmegaSimpleLogger, OmegaLoggingHandler
from omegaml.tests.util import OmegaTestMixin


class OmegaLoggingTests(OmegaTestMixin, unittest.TestCase):
    def setUp(self):
        self.om = Omega()
        self.clean()
        self._reset_log()

    def tearDown(self):
        self.assertTrue(all(m.name is not None for m in self.om.datasets.list(raw=True)))

    def _reset_log(self):
        self.om.datasets.drop('.omega/logs', force=True)

    def test_simple_logger(self):
        logger = self.om.logger
        self.assertIsInstance(logger, OmegaSimpleLogger)
        logger.setLevel('DEBUG')
        logger.info('info message')
        logger.error('error message')
        logger.debug('debug message')
        logger.warning('warning message')
        logger.critical('critical message')
        for level in OmegaSimpleLogger.levels:
            expected = 1 if level != 'QUIET' else 0
            df = logger.dataset.get(levelname=level)
            self.assertTrue(len(df) == expected, 'level {} did not have a record'.format(level))

    def test_simple_logger_level(self):
        logger = self.om.logger
        self.assertIsInstance(logger, OmegaSimpleLogger)
        logger.setLevel('INFO')
        logger.info('info message')
        logger.debug('error message')
        df = logger.dataset.get(levelname='DEBUG')
        self.assertTrue(len(df) == 0)
        df = logger.dataset.get(levelname='INFO')
        self.assertTrue(len(df) == 1)

    def test_simple_logger_quiet(self):
        logger = self.om.logger
        self.assertIsInstance(logger, OmegaSimpleLogger)
        logger.info('initialize') # need this to initialize
        logger.setLevel('QUIET')
        logger.error('info message')
        logger.debug('error message')
        df = logger.dataset.get(levelname='INFO')
        self.assertTrue(len(df) == 1)
        df = logger.dataset.get(levelname='DEBUG')
        self.assertTrue(len(df) == 0)
        df = logger.dataset.get(levelname='ERROR')
        self.assertTrue(len(df) == 0)

    def test_loghandler(self):
        pylogger = logging.getLogger()
        omlogger = self.om.logger
        handler = OmegaLoggingHandler.setup(logger=pylogger)
        pylogger.info('test')
        df = omlogger.dataset.get(levelname='INFO')
        self.assertTrue(len(df) == 1) # includes initialization message
        pylogger.handlers.remove(handler)







# -- FILE: features/environment.py
# CONTAINS: Browser fixture setup and teardown
import os

import nbformat
from behave import fixture, use_fixture
from selenium.webdriver import ChromeOptions
from splinter.browser import Browser

from omegaml import settings
from omegaml.tests.features.util import istrue
from omegaml.tests.util import clear_om


@fixture
def splinter_browser(context):
    headless = istrue(os.environ.get('CHROME_HEADLESS'))
    screenshot_path = os.environ.get('CHROME_SCREENSHOTS', '/tmp/screenshots')
    os.makedirs(screenshot_path, exist_ok=True)
    options = None
    if headless:
        print("Running headless, debug at http://localhost:9222")
        options = ChromeOptions()
        options.add_argument('--no-sandbox')
        options.add_argument('--headless')
        options.add_argument('--disable-gpu')
        options.add_argument('--remote-debugging-port=9222')
        options.add_argument('--remote-debugging-address=0.0.0.0')
        options.add_argument('--disable-dev-shm-usage')
        options.add_argument('--disable-extensions')
    context.browser = Browser('chrome', options=options)
    context.browser.driver.set_window_size(1024, 768)
    context.screenshot_path = screenshot_path
    yield context.browser
    context.browser.quit()


def before_all(context):
    # setup omegaml
    import omegaml as om
    use_fixture(splinter_browser, context)
    # set url and admin password
    context.web_url = os.environ.get('OMEGA_URL', 'http://localhost:5000')
    context.jynb_url = os.environ.get('JUPYTER_URL', 'http://localhost:8888')
    # setup environment
    context.debug = os.environ.get('BEHAVE_DEBUG', False)
    defaults = settings()
    context.om = om.setup()
    context.nbfiles = os.environ.get('BEHAVE_NBFILES', './docs/source/nb')

def before_scenario(context, scenario):
    # FIXME we do this because context.feature is set dynamically in EE testing
    context.feature.jynb_url = context.jynb_url

def after_step(context, step):
    context.screenshotfn = os.path.join(context.screenshot_path, step.name + '.png')
    context.browser.screenshot(context.screenshotfn)
    if context.debug and step.status == "failed":
        # -- ENTER DEBUGGER: Zoom in on failure location.
        # NOTE: Use IPython debugger, same for pdb (basic python debugger).
        import ipdb
        ipdb.post_mortem(step.exc_traceback)
    # also copy notebooks for debugging
    for nbname in context.om.jobs.list():
        try:
            nb = context.om.jobs.get(nbname)
            if nb is not None:
                outname = os.path.join(context.screenshot_path, nbname)
                dirname = os.path.dirname(outname)
                os.makedirs(dirname, exist_ok=True)
                nbformat.write(nb, outname)
        except Exception as e:
            print("WARNING could not write {nbname} to {outname}, got {e}".format(**locals()))


def after_scenario(context, scenario):
    try:
        if hasattr(context, 'om'):
            clear_om(context.om)
        if hasattr(context.feature, 'om'):
            clear_om(context.feature.om)
        if hasattr(scenario, 'om'):
            clear_om(scenario.om)
    except:
        if context.debug:
            import ipdb
            ipdb.post_mortem()




from .tfestimatormodel import TFEstimatorModelBackend, TFEstimatorModel
from .tfkeras import TensorflowKerasBackend
from .tfsavedmodel import TensorflowSavedModelPredictor, TensorflowSavedModelBackend, ServingInput
from .tfkerassavedmodel import  TensorflowKerasSavedModelBackend



from unittest import TestCase, skipIf

from omegaml import Omega
from omegaml.backends.tensorflow.tfsavedmodel import TensorflowSavedModelBackend, TensorflowSavedModelPredictor
from omegaml.tests.util import OmegaTestMixin, tf_in_eager_execution


class TensorflowSavedModelBackendTests(OmegaTestMixin, TestCase):
    def setUp(self):
        self.om = Omega()
        self.om.models.register_backend(TensorflowSavedModelBackend.KIND, TensorflowSavedModelBackend)
        self.clean()

    def _build_model(self):
        # build a dummy model for testing. does not need to make sense
        import tensorflow as tf
        keras = tf.keras
        Sequential = keras.models.Sequential
        Dense = keras.layers.Dense
        Dropout = keras.layers.Dropout
        SGD = keras.optimizers.SGD

        # Generate dummy data
        import numpy as np
        x_train = np.random.random((1000, 20))
        y_train = keras.utils.to_categorical(np.random.randint(10, size=(1000, 1)), num_classes=10)
        x_test = np.random.random((100, 20))
        y_test = keras.utils.to_categorical(np.random.randint(10, size=(100, 1)), num_classes=10)

        model = Sequential()
        # Dense(64) is a fully-connected layer with 64 hidden units.
        # in the first layer, you must specify the expected input data shape:
        # here, 20-dimensional vectors.
        model.add(Dense(64, activation='relu', input_dim=20, name='X'))
        model.add(Dropout(0.5))
        model.add(Dense(64, activation='relu'))
        model.add(Dropout(0.5))
        model.add(Dense(10, activation='softmax'))
        sgd = SGD(lr=0.01, decay=1e-6, momentum=0.9, nesterov=True)
        model.compile(loss='categorical_crossentropy',
                      optimizer=sgd,
                      metrics=['accuracy'])

        # https://www.tensorflow.org/guide/estimators
        est_model = tf.keras.estimator.model_to_estimator(keras_model=model)
        train_input_fn = tf.estimator.inputs.numpy_input_fn(
            x={"X_input": x_train},
            y=y_train,
            num_epochs=1,
            shuffle=False)

        est_model.train(train_input_fn)
        return est_model

    @skipIf(tf_in_eager_execution(), "cannot run in eager mode")
    def test_save_load_tf_example(self):
        import tensorflow as tf
        import numpy as np

        om = self.om
        model = self._build_model()

        # https://www.tensorflow.org/guide/saved_model#prepare_serving_inputs
        default_batch_size = 1
        feature_spec = {'X_input': tf.FixedLenFeature(dtype=np.float32, shape=(20,))}

        def serving_input_receiver_fn():
            """An input receiver that expects a serialized tf.Example."""
            serialized_tf_example = tf.placeholder(dtype=tf.string,
                                                   shape=[default_batch_size],
                                                   name='X_input')
            receiver_tensors = {'X_input': serialized_tf_example}
            features = tf.parse_example(serialized_tf_example, feature_spec)
            return tf.estimator.export.ServingInputReceiver(features, receiver_tensors)

        x_test = np.random.random((100, 20))
        def input_fn():
            X = tf.data.Dataset.from_tensor_slices(x_test)
            return X.batch(1)
        yhat = [v for v in model.predict(input_fn=input_fn)]
        om.models.put(model, 'estimator-savedmodel',
                      serving_input_fn=serving_input_receiver_fn)
        self.assertIn('estimator-savedmodel', om.models.list())
        model_ = om.models.get('estimator-savedmodel')
        self.assertIsInstance(model_, TensorflowSavedModelPredictor)
        for i in range(x_test.shape[0]):
            example = tf.train.Example(features=tf.train.Features(feature={
                'X_input': tf.train.Feature(float_list=tf.train.FloatList(value=x_test[i, :]))
            }))
            yhat_ = model_.predict([example.SerializeToString()])
            self.assertTrue(np.allclose(yhat_, yhat[i][model_.output_names[0]]))

    @skipIf(tf_in_eager_execution(), "cannot run in eager mode")
    def test_save_load_native(self):
        import tensorflow as tf
        import numpy as np

        om = self.om
        model = self._build_model()

        # https://www.tensorflow.org/guide/saved_model#prepare_serving_inputs
        default_batch_size = 1

        def serving_input_receiver_fn():
            placeholder = tf.placeholder(dtype=np.float32,
                                         shape=(default_batch_size, 20),
                                         name='X_input')
            receiver_tensors = {'X_input': placeholder}
            features = {'X_input': placeholder}
            return tf.estimator.export.ServingInputReceiver(features, receiver_tensors)

        x_test = np.random.random((100, 20))
        def input_fn():
            X = tf.data.Dataset.from_tensor_slices(x_test)
            return X.batch(1)
        yhat = [v for v in model.predict(input_fn=input_fn)]
        om.models.put(model, 'estimator-savedmodel',
                      serving_input_fn=serving_input_receiver_fn)
        self.assertIn('estimator-savedmodel', om.models.list())
        model_ = om.models.get('estimator-savedmodel')
        self.assertIsInstance(model_, TensorflowSavedModelPredictor)
        for i in range(x_test.shape[0]):
            yhat_ = model_.predict([x_test[i]])
            self.assertTrue(np.allclose(yhat_, yhat[i][model_.output_names[0]]))

    def test_prediction_restapi(self):
        import tensorflow as tf
        import numpy as np

        om = self.om
        model = self._build_model()

        default_batch_size = 1
        def serving_input_receiver_fn():
            placeholder = tf.placeholder(dtype=np.float32,
                                         shape=(default_batch_size, 20),
                                         name='X_input')
            receiver_tensors = {'X_input': placeholder}
            features = {'X_input': placeholder}
            return tf.estimator.export.ServingInputReceiver(features, receiver_tensors)






class RequestsLikeTestClient(object):
    # inspired by https://stackoverflow.com/a/41151251
    def __init__(self, app):
        self.client = app.test_client()
        self.headers = None

    def make_client_kwargs(self, json=None, auth=None, headers=None, **kwargs):
        self.headers = headers or {}
        if auth:
            auth(self)
        if json:
            from json import dumps
            self.headers.update({
                'Accept': 'application/json',
                'Content-Type': 'application/json'
            })
            kwargs.update(data=dumps(json))
        kwargs.update(headers=self.headers)
        return kwargs

    def wrapper(method):
        def inner(self, *args, **kwargs):
            kwargs = self.make_client_kwargs(**kwargs)
            return getattr(self.client, method)(*args, **kwargs)
        return inner

    get = wrapper('get')
    post = wrapper('post')
    put = wrapper('put')
    delete = wrapper('delete')




#!/usr/bin/env python
from distutils.core import setup

setup(name='helloworld', version='1.0', description='simple omegaml hello world script', author='omegaml',
      author_email='info@omegaml.io', url='http://omegaml.io', packages=['helloworld'], )



# Configuration file for jupyter-notebook.
import os
import sys


def setup():
    # avoid IDE error
    global c
    return c


c = setup()

# ------------------------------------------------------------------------------
# Application(SingletonConfigurable) configuration
# ------------------------------------------------------------------------------

# This is an application.

# The date format used by logging formatters for %(asctime)s
# c.Application.log_datefmt = '%Y-%m-%d %H:%M:%S'

# The Logging format template
# c.Application.log_format = '[%(name)s]%(highlevel)s %(message)s'

# Set the log level by value or name.
# c.Application.log_level = 30

# ------------------------------------------------------------------------------
# JupyterApp(Application) configuration
# ------------------------------------------------------------------------------

# Base class for Jupyter applications

# Answer yes to any prompts.
# c.JupyterApp.answer_yes = False

# Full path of a config file.
# c.JupyterApp.config_file = ''

# Specify a config file to load.
# c.JupyterApp.config_file_name = ''

# Generate default config file.
# c.JupyterApp.generate_config = False

# ------------------------------------------------------------------------------
# NotebookApp(JupyterApp) configuration
# ------------------------------------------------------------------------------

# Set the Access-Control-Allow-Credentials: true header
# c.NotebookApp.allow_credentials = False

# Set the Access-Control-Allow-Origin header
#
#  Use '*' to allow any origin to access your server.
#
#  Takes precedence over allow_origin_pat.
# c.NotebookApp.allow_origin = ''

# Use a regular expression for the Access-Control-Allow-Origin header
#
#  Requests from an origin matching the expression will get replies with:
#
#      Access-Control-Allow-Origin: origin
#
#  where `origin` is the origin of the request.
#
#  Ignored if allow_origin is set.
# c.NotebookApp.allow_origin_pat = ''

# Whether to allow the user to run the notebook as root.
c.NotebookApp.allow_root = True if os.environ.get('JY_ALLOW_ROOT') else False

# DEPRECATED use base_url
# c.NotebookApp.base_project_url = '/'

# The base URL for the notebook server.
#
#  Leading and trailing slashes can be omitted, and will automatically be added.
# c.NotebookApp.base_url = '/'

# Specify what command to use to invoke a web browser when opening the notebook.
#  If not specified, the default browser will be determined by the `webbrowser`
#  standard library module, which allows setting of the BROWSER environment
#  variable to override it.
# c.NotebookApp.browser = ''

# The full path to an SSL/TLS certificate file.
# c.NotebookApp.certfile = ''

# The full path to a certificate authority certificate for SSL/TLS client
#  authentication.
# c.NotebookApp.client_ca = ''

# The config manager class to use
# c.NotebookApp.config_manager_class = 'notebook.services.config.manager.ConfigManager'

# The notebook manager class to use.
# c.NotebookApp.contents_manager_class = 'notebook.services.contents.largefilemanager.LargeFileManager'

# Extra keyword arguments to pass to `set_secure_cookie`. See tornado's
#  set_secure_cookie docs for details.
# c.NotebookApp.cookie_options = {}

# The random bytes used to secure cookies. By default this is a new random
#  number every time you start the Notebook. Set it to a value in a config file
#  to enable logins to persist across server sessions.
#
#  Note: Cookie secrets should be kept private, do not share config files with
#  cookie_secret stored in plaintext (you can read the value from a file).
# c.NotebookApp.cookie_secret = b''

# The file where the cookie secret is stored.
# c.NotebookApp.cookie_secret_file = ''

# The default URL to redirect to from `/`
# c.NotebookApp.default_url = '/tree'

# Disable cross-site-request-forgery protection
#
#  Jupyter notebook 4.3.1 introduces protection from cross-site request
#  forgeries, requiring API requests to either:
#
#  - originate from pages served by this server (validated with XSRF cookie and
#  token), or - authenticate with a token
#
#  Some anonymous compute resources still desire the ability to run code,
#  completely without authentication. These services can disable all
#  authentication and security checks, with the full knowledge of what that
#  implies.
# c.NotebookApp.disable_check_xsrf = False

# Whether to enable MathJax for typesetting math/TeX
#
#  MathJax is the javascript library Jupyter uses to render math/LaTeX. It is
#  very large, so you may want to disable it if you have a slow internet
#  connection, or for offline use of the notebook.
#
#  When disabled, equations etc. will appear as their untransformed TeX source.
# c.NotebookApp.enable_mathjax = True

# extra paths to look for Javascript notebook extensions
# c.NotebookApp.extra_nbextensions_path = []

# Extra paths to search for serving static files.
#
#  This allows adding javascript/css to be available from the notebook server
#  machine, or overriding individual files in the IPython
# c.NotebookApp.extra_static_paths = []

# Extra paths to search for serving jinja templates.
#
#  Can be used to override templates from notebook.templates.
# c.NotebookApp.extra_template_paths = []

##
# c.NotebookApp.file_to_run = ''

# Deprecated: Use minified JS file or not, mainly use during dev to avoid JS
#  recompilation
# c.NotebookApp.ignore_minified_js = False

# (bytes/sec) Maximum rate at which messages can be sent on iopub before they
#  are limited.
# c.NotebookApp.iopub_data_rate_limit = 1000000

# (msgs/sec) Maximum rate at which messages can be sent on iopub before they are
#  limited.
# c.NotebookApp.iopub_msg_rate_limit = 1000

# The IP address the notebook server will listen on.
# c.NotebookApp.ip = 'localhost'

# Supply extra arguments that will be passed to Jinja environment.
# c.NotebookApp.jinja_environment_options = {}

# Extra variables to supply to jinja templates when rendering.
# c.NotebookApp.jinja_template_vars = {}

# The kernel manager class to use.
# c.NotebookApp.kernel_manager_class = 'notebook.services.kernels.kernelmanager.MappingKernelManager'

# The kernel spec manager class to use. Should be a subclass of
#  `jupyter_client.kernelspec.KernelSpecManager`.
#
#  The Api of KernelSpecManager is provisional and might change without warning
#  between this version of Jupyter and the next stable one.
# c.NotebookApp.kernel_spec_manager_class = 'jupyter_client.kernelspec.KernelSpecManager'

# The full path to a private key file for usage with SSL/TLS.
# c.NotebookApp.keyfile = ''

# The login handler class to use.
# c.NotebookApp.login_handler_class = 'notebook.auth.login.LoginHandler'

# The logout handler class to use.
# c.NotebookApp.logout_handler_class = 'notebook.auth.logout.LogoutHandler'

# The MathJax.js configuration file that is to be used.
# c.NotebookApp.mathjax_config = 'TeX-AMS-MML_HTMLorMML-full,Safe'

# A custom url for MathJax.js. Should be in the form of a case-sensitive url to
#  MathJax, for example:  /static/components/MathJax/MathJax.js
# c.NotebookApp.mathjax_url = ''

# Dict of Python modules to load as notebook server extensions.Entry values can
#  be used to enable and disable the loading ofthe extensions. The extensions
#  will be loaded in alphabetical order.
# c.NotebookApp.nbserver_extensions = {}

# The directory to use for notebooks and kernels.
c.NotebookApp.notebook_dir = os.getcwd()

# Whether to open in a browser after starting. The specific browser used is
#  platform dependent and determined by the python standard library `webbrowser`
#  module, unless it is overridden using the --browser (NotebookApp.browser)
#  configuration option.
# c.NotebookApp.open_browser = True

# Hashed password to use for web authentication.
#
#  To generate, type in a python/IPython shell:
#
#    from notebook.auth import passwd; passwd()
#
#  The string should be of the form type:salt:hashed-password.
if 'JUPYTER_PASSWORD' in os.environ:
    # unless it is actually a value in environ we must not set the attribute, as None is not a valid value
    # if the value is not set a token is generated which is what we want in this case
    c.NotebookApp.password = os.environ.get('JUPYTER_PASSWORD')

# Forces users to use a password for the Notebook server. This is useful in a
#  multi user environment, for instance when everybody in the LAN can access each
#  other's machine though ssh.
#
#  In such a case, server the notebook server on localhost is not secure since
#  any user can connect to the notebook server via ssh.
c.NotebookApp.password_required = 'JUPYTER_PASSWORD' in os.environ

# The port the notebook server will listen on.
# c.NotebookApp.port = 8888

# The number of additional ports to try if the specified port is not available.
# c.NotebookApp.port_retries = 50

# DISABLED: use %pylab or %matplotlib in the notebook to enable matplotlib.
# c.NotebookApp.pylab = 'disabled'

# (sec) Time window used to  check the message and data rate limits.
# c.NotebookApp.rate_limit_window = 3

# Reraise exceptions encountered loading server extensions?
# c.NotebookApp.reraise_server_extension_failures = False

# DEPRECATED use the nbserver_extensions dict instead
# c.NotebookApp.server_extensions = []

# The session manager class to use.
# c.NotebookApp.session_manager_class = 'notebook.services.sessions.sessionmanager.SessionManager'

# Supply SSL options for the tornado HTTPServer. See the tornado docs for
#  details.
# c.NotebookApp.ssl_options = {}

# Supply overrides for terminado. Currently only supports "shell_command".
# c.NotebookApp.terminado_settings = {}

# Token used for authenticating first-time connections to the server.
#
#  When no password is enabled, the default is to generate a new, random token.
#
#  Setting to an empty string disables authentication altogether, which is NOT
#  RECOMMENDED.
# c.NotebookApp.token = '<generated>'

# Supply overrides for the tornado.web.Application that the Jupyter notebook
#  uses.
# c.NotebookApp.tornado_settings = {}

# Whether to trust or not X-Scheme/X-Forwarded-Proto and X-Real-Ip/X-Forwarded-
#  For headerssent by the upstream reverse proxy. Necessary if the proxy handles
#  SSL
# c.NotebookApp.trust_xheaders = False

# DEPRECATED, use tornado_settings
# c.NotebookApp.webapp_settings = {}

# The base URL for websockets, if it differs from the HTTP server (hint: it
#  almost certainly doesn't).
#
#  Should be in the form of an HTTP origin: ws[s]://hostname[:port]
# c.NotebookApp.websocket_url = ''

# ------------------------------------------------------------------------------
# ConnectionFileMixin(LoggingConfigurable) configuration
# ------------------------------------------------------------------------------

# Mixin for configurable classes that work with connection files

# JSON file in which to store connection info [default: kernel-<pid>.json]
#
#  This file will contain the IP, ports, and authentication key needed to connect
#  clients to this kernel. By default, this file will be created in the security
#  dir of the current profile, but can be specified by absolute path.
# c.ConnectionFileMixin.connection_file = ''

# set the control (ROUTER) port [default: random]
# c.ConnectionFileMixin.control_port = 0

# set the heartbeat port [default: random]
# c.ConnectionFileMixin.hb_port = 0

# set the iopub (PUB) port [default: random]
# c.ConnectionFileMixin.iopub_port = 0

# Set the kernel's IP address [default localhost]. If the IP address is
#  something other than localhost, then Consoles on other machines will be able
#  to connect to the Kernel, so be careful!
# c.ConnectionFileMixin.ip = ''

# set the shell (ROUTER) port [default: random]
# c.ConnectionFileMixin.shell_port = 0

# set the stdin (ROUTER) port [default: random]
# c.ConnectionFileMixin.stdin_port = 0

##
# c.ConnectionFileMixin.transport = 'tcp'

# ------------------------------------------------------------------------------
# KernelManager(ConnectionFileMixin) configuration
# ------------------------------------------------------------------------------

# Manages a single kernel in a subprocess on this host.
#
#  This version starts kernels with Popen.

# Should we autorestart the kernel if it dies.
# c.KernelManager.autorestart = True

# DEPRECATED: Use kernel_name instead.
#
#  The Popen Command to launch the kernel. Override this if you have a custom
#  kernel. If kernel_cmd is specified in a configuration file, Jupyter does not
#  pass any arguments to the kernel, because it cannot make any assumptions about
#  the arguments that the kernel understands. In particular, this means that the
#  kernel does not receive the option --debug if it given on the Jupyter command
#  line.
# c.KernelManager.kernel_cmd = []

# Time to wait for a kernel to terminate before killing it, in seconds.
# c.KernelManager.shutdown_wait_time = 5.0

# ------------------------------------------------------------------------------
# Session(Configurable) configuration
# ------------------------------------------------------------------------------

# Object for handling serialization and sending of messages.
#
#  The Session object handles building messages and sending them with ZMQ sockets
#  or ZMQStream objects.  Objects can communicate with each other over the
#  network via Session objects, and only need to work with the dict-based IPython
#  message spec. The Session will handle serialization/deserialization, security,
#  and metadata.
#
#  Sessions support configurable serialization via packer/unpacker traits, and
#  signing with HMAC digests via the key/keyfile traits.
#
#  Parameters ----------
#
#  debug : bool
#      whether to trigger extra debugging statements
#  packer/unpacker : str : 'json', 'pickle' or import_string
#      importstrings for methods to serialize message parts.  If just
#      'json' or 'pickle', predefined JSON and pickle packers will be used.
#      Otherwise, the entire importstring must be used.
#
#      The functions must accept at least valid JSON input, and output *bytes*.
#
#      For example, to use msgpack:
#      packer = 'msgpack.packb', unpacker='msgpack.unpackb'
#  pack/unpack : callables
#      You can also set the pack/unpack callables for serialization directly.
#  session : bytes
#      the ID of this Session object.  The default is to generate a new UUID.
#  username : unicode
#      username added to message headers.  The default is to ask the OS.
#  key : bytes
#      The key used to initialize an HMAC signature.  If unset, messages
#      will not be signed or checked.
#  keyfile : filepath
#      The file containing a key.  If this is set, `key` will be initialized
#      to the contents of the file.

# Threshold (in bytes) beyond which an object's buffer should be extracted to
#  avoid pickling.
# c.Session.buffer_threshold = 1024

# Whether to check PID to protect against calls after fork.
#
#  This check can be disabled if fork-safety is handled elsewhere.
# c.Session.check_pid = True

# Threshold (in bytes) beyond which a buffer should be sent without copying.
# c.Session.copy_threshold = 65536

# Debug output in the Session
c.Session.debug = False

# The maximum number of digests to remember.
#
#  The digest history will be culled when it exceeds this value.
# c.Session.digest_history_size = 65536

# The maximum number of items for a container to be introspected for custom
#  serialization. Containers larger than this are pickled outright.
# c.Session.item_threshold = 64

# execution key, for signing messages.
# c.Session.key = b''

# path to file containing execution key.
# c.Session.keyfile = ''

# Metadata dictionary, which serves as the default top-level metadata dict for
#  each message.
# c.Session.metadata = {}

# The name of the packer for serializing messages. Should be one of 'json',
#  'pickle', or an import name for a custom callable serializer.
# c.Session.packer = 'json'

# The UUID identifying this session.
# c.Session.session = ''

# The digest scheme used to construct the message signatures. Must have the form
#  'hmac-HASH'.
# c.Session.signature_scheme = 'hmac-sha256'

# The name of the unpacker for unserializing messages. Only used with custom
#  functions for `packer`.
# c.Session.unpacker = 'json'

# Username for the Session. Default is your system username.
# c.Session.username = 'patrick'

# ------------------------------------------------------------------------------
# MultiKernelManager(LoggingConfigurable) configuration
# ------------------------------------------------------------------------------

# A class for managing multiple kernels.

# The name of the default kernel to start
# c.MultiKernelManager.default_kernel_name = 'python3'

# The kernel manager class.  This is configurable to allow subclassing of the
#  KernelManager for customized behavior.
# c.MultiKernelManager.kernel_manager_class = 'jupyter_client.ioloop.IOLoopKernelManager'

# ------------------------------------------------------------------------------
# MappingKernelManager(MultiKernelManager) configuration
# ------------------------------------------------------------------------------

# A KernelManager that handles notebook mapping and HTTP error handling

##
# c.MappingKernelManager.root_dir = ''

# ------------------------------------------------------------------------------
# ContentsManager(LoggingConfigurable) configuration
# ------------------------------------------------------------------------------

# Base class for serving files and directories.
#
#  This serves any text or binary file, as well as directories, with special
#  handling for JSON notebook documents.
#
#  Most APIs take a path argument, which is always an API-style unicode path, and
#  always refers to a directory.
#
#  - unicode, not url-escaped
#  - '/'-separated
#  - leading and trailing '/' will be stripped
#  - if unspecified, path defaults to '',
#    indicating the root path.

##
# c.ContentsManager.checkpoints = None

##
# c.ContentsManager.checkpoints_class = 'notebook.services.contents.checkpoints.Checkpoints'

##
# c.ContentsManager.checkpoints_kwargs = {}

# Glob patterns to hide in file and directory listings.
# c.ContentsManager.hide_globs = ['__pycache__', '*.pyc', '*.pyo', '.DS_Store', '*.so', '*.dylib', '*~']

# Python callable or importstring thereof
#
#  To be called on a contents model prior to save.
#
#  This can be used to process the structure, such as removing notebook outputs
#  or other side effects that should not be saved.
#
#  It will be called as (all arguments passed by keyword)::
#
#      hook(path=path, model=model, contents_manager=self)
#
#  - model: the model to be saved. Includes file contents.
#    Modifying this dict will affect the file that is stored.
#  - path: the API path of the save destination
#  - contents_manager: this ContentsManager instance
# c.ContentsManager.pre_save_hook = None

##
# c.ContentsManager.root_dir = '/'

# The base name used when creating untitled directories.
# c.ContentsManager.untitled_directory = 'Untitled Folder'

# The base name used when creating untitled files.
# c.ContentsManager.untitled_file = 'untitled'

# The base name used when creating untitled notebooks.
# c.ContentsManager.untitled_notebook = 'Untitled'

# ------------------------------------------------------------------------------
# FileManagerMixin(Configurable) configuration
# ------------------------------------------------------------------------------

# Mixin for ContentsAPI classes that interact with the filesystem.
#
#  Provides facilities for reading, writing, and copying both notebooks and
#  generic files.
#
#  Shared by FileContentsManager and FileCheckpoints.
#
#  Note ---- Classes using this mixin must provide the following attributes:
#
#  root_dir : unicode
#      A directory against against which API-style paths are to be resolved.
#
#  log : logging.Logger

# By default notebooks are saved on disk on a temporary file and then if
#  succefully written, it replaces the old ones. This procedure, namely
#  'atomic_writing', causes some bugs on file system whitout operation order
#  enforcement (like some networked fs). If set to False, the new notebook is
#  written directly on the old one which could fail (eg: full filesystem or quota
#  )
# c.FileManagerMixin.use_atomic_writing = True

# ------------------------------------------------------------------------------
# FileContentsManager(FileManagerMixin,ContentsManager) configuration
# ------------------------------------------------------------------------------

# Python callable or importstring thereof
#
#  to be called on the path of a file just saved.
#
#  This can be used to process the file on disk, such as converting the notebook
#  to a script or HTML via nbconvert.
#
#  It will be called as (all arguments passed by keyword)::
#
#      hook(os_path=os_path, model=model, contents_manager=instance)
#
#  - path: the filesystem path to the file just written - model: the model
#  representing the file - contents_manager: this ContentsManager instance
# c.FileContentsManager.post_save_hook = None

##
# c.FileContentsManager.root_dir = ''

# DEPRECATED, use post_save_hook. Will be removed in Notebook 5.0
# c.FileContentsManager.save_script = False

# ------------------------------------------------------------------------------
# NotebookNotary(LoggingConfigurable) configuration
# ------------------------------------------------------------------------------

# A class for computing and verifying notebook signatures.

# The hashing algorithm used to sign notebooks.
# c.NotebookNotary.algorithm = 'sha256'

# The sqlite file in which to store notebook signatures. By default, this will
#  be in your Jupyter data directory. You can set it to ':memory:' to disable
#  sqlite writing to the filesystem.
# c.NotebookNotary.db_file = ''

# The secret key with which notebooks are signed.
# c.NotebookNotary.secret = b''

# The file where the secret key is stored.
# c.NotebookNotary.secret_file = ''

# A callable returning the storage backend for notebook signatures. The default
#  uses an SQLite database.
# c.NotebookNotary.store_factory = traitlets.Undefined

# ------------------------------------------------------------------------------
# KernelSpecManager(LoggingConfigurable) configuration
# ------------------------------------------------------------------------------

# If there is no Python kernelspec registered and the IPython kernel is
#  available, ensure it is added to the spec list.
# c.KernelSpecManager.ensure_native_kernel = True

# The kernel spec class.  This is configurable to allow subclassing of the
#  KernelSpecManager for customized behavior.
# c.KernelSpecManager.kernel_spec_class = 'jupyter_client.kernelspec.KernelSpec'

# Whitelist of allowed kernel names.
#
#  By default, all installed kernels are allowed.
# c.KernelSpecManager.whitelist = set()


# omegaml setup
default_contents_manager = 'omegaml.notebook.omegacontentsmgr.OmegaStoreContentsManager'
contents_manager = os.environ.get('JY_CONTENTS_MANAGER') or default_contents_manager
c.NotebookApp.contents_manager_class = contents_manager

import logging

logger = logging.getLogger(c.NotebookApp.__class__.__name__)
handler = logging.FileHandler('/tmp/jupyter.log')
logger.addHandler(handler)
logger.setLevel(os.environ.get('JY_DEBUG_LEVEL', 'INFO'))

# if we're running inside a container, use /app
if 'OMEGA_ROOT' in os.environ:
    sys.path.insert(0, os.environ['OMEGA_ROOT'])
else:
    sys.path.insert(0, '/app')





from ._version import version
from .store.logging import OmegaSimpleLogger


class Omega(object):
    """
    Client API to omegaml

    Provides the following APIs:

    * :code:`datasets` - access to datasets stored in the cluster
    * :code:`models` - access to models stored in the cluster
    * :code:`runtimes` - access to the cluster compute resources
    * :code:`jobs` - access to jobs stored and executed in the cluster
    * :code:`scripts` - access to lambda modules stored and executed in the cluster

    """

    def __init__(self, defaults=None, mongo_url=None, celeryconf=None, bucket=None,
                 **kwargs):
        """
        Initialize the client API

        Without arguments create the client API according to the user's
        configuration in :code:`~/omegaml/config.yml`.

        Arguments override the user's configuration.

            :param defaults: the DefaultsContext
        :param mongo_url: the fully qualified URI to the mongo database,
        of format :code:`mongodb://user:password@host:port/database`
        :param celeryconf: the celery configuration dictionary
        """
        from omegaml.util import settings
        # avoid circular imports
        from omegaml.notebook.jobs import OmegaJobs
        from omegaml.store import OmegaStore
        # celery and mongo configuration
        self.defaults = defaults or settings()
        self.mongo_url = mongo_url or self.defaults.OMEGA_MONGO_URL
        self.bucket = bucket
        # setup storage locations
        self.models = OmegaStore(mongo_url=self.mongo_url, bucket=bucket, prefix='models/', defaults=self.defaults)
        self.datasets = OmegaStore(mongo_url=self.mongo_url, bucket=bucket, prefix='data/', defaults=self.defaults)
        self._jobdata = OmegaStore(mongo_url=self.mongo_url, bucket=bucket, prefix='jobs/', defaults=self.defaults)
        self.scripts = OmegaStore(mongo_url=self.mongo_url, prefix='scripts/', defaults=self.defaults)
        # runtimes environments
        self.runtime = self._make_runtime(celeryconf)
        self.jobs = OmegaJobs(store=self._jobdata)
        # logger
        self.logger = OmegaSimpleLogger(store=self.datasets, defaults=self.defaults)

    def __repr__(self):
        return 'Omega()'.format()

    def _clone(self, **kwargs):
        return self.__class__(defaults=self.defaults,
                              mongo_url=self.mongo_url,
                              **kwargs)

    def _make_runtime(self, celeryconf):
        from omegaml.runtimes import OmegaRuntime
        return OmegaRuntime(self, bucket=self.bucket, defaults=self.defaults, celeryconf=celeryconf)

    def __getitem__(self, bucket):
        """
        return Omega instance configured for the given bucket

        Args:
            bucket (str): the bucket name. If it does not exist
                  it gets created on first storage of an object.
                  If bucket=None returns self.

        Usage:
            import omegaml as om

            # om is configured on the default bucket
            # om_mybucket will use the same settings, but configured for mybucket
            om_mybucket = om['mybucket']

        Returns:
            Omega instance configured for the given bucket
        """
        if bucket is None or self.bucket == bucket:
            return self
        return self._clone(bucket=bucket)


class OmegaDeferredInstance(object):
    """
    A deferred instance of Omega() that is only instantiated on access

    This is to ensure that module-level imports don't trigger instantiation
    of Omega.
    """

    def __init__(self, base=None, attribute=None):
        self.omega = 'not initialized -- call .setup() or access an attribute'
        self.initialized = False
        self.base = base
        self.attribute = attribute

    def setup(self, mongo_url=None, bucket=None, celeryconf=None):
        from omegaml.client.cloud import setup_from_config
        try:
            omega = setup_from_config()
        except SystemError:
            omega = Omega(mongo_url=None, bucket=bucket, celeryconf=None)
        if not self.initialized:
            self.initialized = True
            self.omega = omega
        return omega

    def __getattr__(self, name):
        if self.base:
            base = getattr(self.base, self.attribute)
            return getattr(base, name)
        if not self.initialized:
            self.setup()
        return getattr(self.omega, name)

    def __getitem__(self, bucket):
        if self.base:
            base = getattr(self.base, self.attribute)
            return base[bucket]
        if not self.initialized:
            self.setup()
        return self.omega[bucket]

    def __repr__(self):
        if self.base:
            return repr(getattr(self.base, self.attribute))
        self.setup()
        return repr(self.omega)


def setup(*args, **kwargs):
    """
    configure and return the omega client instance
    """
    return _om.setup(*args, **kwargs)


# dynamic lookup of Omega instance in a task context
get_omega_for_task = lambda *args, **kwargs: _om.setup(*args, **kwargs)

# default instance
# -- these are deferred instanced that is the actual Omega instance
#    is only created on actual attribute access
_om = OmegaDeferredInstance()
version = version  # ensure we keep imports





from __future__ import absolute_import

from datetime import timedelta
from unittest import TestCase

import gridfs
from nbformat import v4

from omegaml import Omega
from omegaml.notebook.jobs import JobSchedule
from omegaml.util import settings as omegaml_settings


class JobTests(TestCase):

    def setup(self):
        TestCase.setUp(self)

    def tearDown(self):
        TestCase.tearDown(self)
        for fn in self.om.jobs.list():
            self.om.jobs.drop(fn)

    @property
    def om(self):
        om = Omega()
        return om

    @property
    def fs(self):
        om = self.om
        defaults = omegaml_settings()
        fs = om.jobs.get_fs(defaults.OMEGA_NOTEBOOK_COLLECTION)
        return fs

    def test_job_put_get(self):
        """
        test job put and get
        """
        om = self.om
        # create a notebook
        cells = []
        code = "print 'hello'"
        cells.append(v4.new_code_cell(source=code))
        notebook = v4.new_notebook(cells=cells)
        # put the notebook
        meta = om.jobs.put(notebook, 'testjob')
        self.assertEqual(meta.name, 'testjob.ipynb')
        # read it back and see what's in it
        notebook2 = om.jobs.get('testjob')
        self.assertDictEqual(notebook2, notebook)

    def test_job_list(self):
        """
        test job listing
        """
        fs = self.fs
        om = self.om
        # create a notebook
        cells = []
        code = "print 'hello'"
        cells.append(v4.new_code_cell(source=code))
        notebook = v4.new_notebook(cells=cells)
        # put the notebook
        meta = om.jobs.put(notebook, 'testjob')
        self.assertEqual(meta.name, 'testjob.ipynb')
        nb = v4.new_notebook(cells=cells)
        job_list = self.om.jobs.list()
        expected = 'testjob.ipynb'
        self.assertIn(expected, job_list)

    def test_run_job_valid(self):
        """
        test running a valid job
        """
        om = self.om
        # create a notebook
        cells = []
        code = "print('hello')"
        cells.append(v4.new_code_cell(source=code))
        notebook = v4.new_notebook(cells=cells)
        # put the notebook
        meta = om.jobs.put(notebook, 'testjob')
        self.assertEqual(meta.name, 'testjob.ipynb')
        meta_job = om.jobs.run('testjob')
        self.assertIn('job_results', meta_job.attributes)
        self.assertIn('job_runs', meta_job.attributes)
        runs = meta_job.attributes['job_runs']
        results = meta_job.attributes['job_results']
        self.assertEqual(len(runs), 1)
        self.assertEqual(len(results), 1)
        resultnb = results[0]
        self.assertTrue(om.jobs.exists(resultnb))
        self.assertEqual(runs[0]['results'], resultnb)

    def test_run_job_timeout(self):
        """
        test running a valid job
        """
        om = self.om
        # create a long-running notebook
        cells = []
        code = "import time; time.sleep(15)"
        cells.append(v4.new_code_cell(source=code))
        notebook = v4.new_notebook(cells=cells)
        meta = om.jobs.put(notebook, 'testjob')
        # -- execute with default timeout, expect to succeed
        meta_job = om.jobs.run('testjob')
        runs = meta_job.attributes['job_runs']
        results = meta_job.attributes['job_results']
        self.assertIn('job_runs', meta_job.attributes)
        self.assertEqual(len(results), 1)
        resultnb = results[0]
        self.assertTrue(om.jobs.exists(resultnb))
        self.assertEqual(runs[0]['results'], resultnb)
        # -- put the notebook with a timeout less than expected running time
        # -- expect run to fail due to timeout
        om.jobs.drop('testjob', force=True)
        meta = om.jobs.put(notebook, 'testjob')
        meta.kind_meta['ep_kwargs'] = dict(timeout=5)
        meta.save()
        self.assertEqual(meta.name, 'testjob.ipynb')
        meta_job = om.jobs.run('testjob')
        self.assertIn('job_runs', meta_job.attributes)
        runs = meta_job.attributes['job_runs']
        this_run = runs[0]
        self.assertEqual(this_run['status'], 'ERROR')
        self.assertIn('execution timed out', this_run['message'])
        self.assertEqual(len(runs), 1)
        # -- retry with no timeout
        om.jobs.drop('testjob', force=True)
        meta = om.jobs.put(notebook, 'testjob')
        meta.kind_meta['ep_kwargs'] = dict(timeout=None)
        meta.save()
        meta_job = om.jobs.run('testjob')
        runs = meta_job.attributes['job_runs']
        results = meta_job.attributes['job_results']
        self.assertIn('job_runs', meta_job.attributes)
        self.assertEqual(len(results), 1)
        resultnb = results[0]
        self.assertTrue(om.jobs.exists(resultnb))
        self.assertEqual(runs[0]['results'], resultnb)

    def test_run_job_invalid(self):
        """
        test running an invalid job
        """
        fs = self.fs
        om = self.om
        # create a notebook
        cells = []
        code = "INVALID PYTHON CODE"
        cells.append(v4.new_code_cell(source=code))
        notebook = v4.new_notebook(cells=cells)
        # put the notebook
        meta = om.jobs.put(notebook, 'testjob')
        self.assertEqual(meta.name, 'testjob.ipynb')
        nb = v4.new_notebook(cells=cells)
        meta_job = om.jobs.run('testjob')
        runs = meta_job.attributes['job_runs']
        self.assertEqual(len(runs), 1)
        self.assertEqual('ERROR', runs[0]['status'])

    def test_run_nonexistent_job(self):
        om = self.om
        self.assertRaises(
            gridfs.errors.NoFile, om.jobs.run_notebook, 'dummys.ipynb')

    def test_scheduled_job_with_omegaml_block(self):
        om = self.om
        cells = []
        conf = """
        # omega-ml:
        #   run-at: "*/5 * * * *"
        """.strip()
        cmd = "print('hello')"
        cells.append(v4.new_code_cell(source=conf))
        cells.append(v4.new_code_cell(source=cmd))
        notebook = v4.new_notebook(cells=cells)
        # check we have a valid configuration object
        meta = om.jobs.put(notebook, 'testjob')
        self._check_scheduled_job()

    def test_scheduled_job_with_run_at_schedule(self):
        om = self.om
        cells = []
        conf = """
        # run-at: "*/5 * * * *"
        """.strip()
        cmd = "print('hello')"
        cells.append(v4.new_code_cell(source=conf))
        cells.append(v4.new_code_cell(source=cmd))
        notebook = v4.new_notebook(cells=cells)
        # check we have a valid configuration object
        meta = om.jobs.put(notebook, 'testjob')
        self._check_scheduled_job()

    def test_scheduled_job_with_cron_schedule(self):
        om = self.om
        cells = []
        conf = """
        # run-at: "*/5 * * * *"
        """.strip()
        cmd = "print('hello')"
        cells.append(v4.new_code_cell(source=conf))
        cells.append(v4.new_code_cell(source=cmd))
        notebook = v4.new_notebook(cells=cells)
        # check we have a valid configuration object
        meta = om.jobs.put(notebook, 'testjob')
        self._check_scheduled_job()

    def test_scheduled_job_with_nlp_schedule(self):
        om = self.om
        cells = []
        conf = """
        # schedule: daily, at 06:00
        """.strip()
        cmd = "print('hello')"
        cells.append(v4.new_code_cell(source=conf))
        cells.append(v4.new_code_cell(source=cmd))
        notebook = v4.new_notebook(cells=cells)
        om.jobs.put(notebook, 'testjob')
        self._check_scheduled_job()

    def test_scheduled_not_duplicated(self):
        om = self.om
        cells = []
        conf = """
                # schedule: daily, at 06:00
                """.strip()
        cmd = "print('hello')"
        cells.append(v4.new_code_cell(source=conf))
        cells.append(v4.new_code_cell(source=cmd))
        notebook = v4.new_notebook(cells=cells)
        om.jobs.put(notebook, 'testjob')
        self._check_scheduled_job(autorun=False, reschedule=False)
        meta = om.jobs.metadata('testjob')
        trigger = meta.attributes['triggers']
        self.assertEqual(len(trigger), 1)

    def test_scheduled_results_not_rescheduled(self):
        om = self.om
        cells = []
        conf = """
                        # schedule: daily, at 06:00
                        """.strip()
        cmd = "print('hello')"
        cells.append(v4.new_code_cell(source=conf))
        cells.append(v4.new_code_cell(source=cmd))
        notebook = v4.new_notebook(cells=cells)
        om.jobs.put(notebook, 'testjob')
        self._check_scheduled_job(autorun=True, reschedule=False)
        meta = om.jobs.metadata('testjob')
        runs = meta.attributes['job_runs']
        result_name = runs[-1]['results']
        meta_result = om.jobs.metadata(result_name)
        self.assertNotIn('triggers', meta_result.attributes)

    def _check_scheduled_job(self, autorun=True, reschedule=False):
        om = self.om
        meta = om.jobs.metadata('testjob')
        config = om.jobs.get_notebook_config('testjob')
        self.assertIn('run-at', config)
        self.assertIn('config', meta.attributes)
        self.assertIn('run-at', meta.attributes['config'])
        # check the job was scheduled
        self.assertIn('triggers', meta.attributes)
        trigger = meta.attributes['triggers'][-1]
        self.assertEqual(trigger['status'], 'PENDING')

        # run it as scheduled, check it is ok
        def get_trigger(event=None):
            # get last trigger or specified by event
            meta = om.jobs.metadata('testjob')
            triggers = meta.attributes['triggers']
            if not event:
                trigger = triggers[-1]
            else:
                trigger = [t for t in triggers if t['event'] == event][0]
            return trigger

        def assert_pending(event=None):
            trigger = get_trigger(event)
            self.assertEqual(trigger['status'], 'PENDING')

        def assert_ok(event=None):
            trigger = get_trigger(event)
            self.assertEqual(trigger['status'], 'OK')

        assert_pending()
        if autorun:
            # -- run by the periodic task. note we pass now= as to simulate a time
            kwargs = dict(now=trigger['run-at'])
            om.runtime.task('omegaml.notebook.tasks.execute_scripts').apply_async(kwargs=kwargs).get()
            assert_ok(event=trigger['event'])
            # -- it should be pending again
            assert_pending()
            # execute the scheduled job by event name
            trigger = get_trigger()
            om.runtime.job('testjob').run(event=trigger['event'])
            # the last run should be ok, and there should not be a new pending event
            # since we did not reschedule
            assert_ok()
            with self.assertRaises(AssertionError):
                assert_pending()
        if reschedule:
            # reschedule explicit and check we have a pending event
            next_time = trigger['run-at'] + timedelta(minutes=2)
            om.runtime.job('testjob').schedule(run_at='daily, at 07:00').get()
            assert_pending()

    def test_schedule_triggers_by_api(self):
        om = self.om
        cells = []
        cmd = "print('hello')"
        cells.append(v4.new_code_cell(source=cmd))
        notebook = v4.new_notebook(cells=cells)
        # check we have a valid configuration object
        om.jobs.put(notebook, 'testjob')
        schedule = om.jobs.Schedule(weekday='mon-fri', at='06:00')
        om.jobs.schedule('testjob', run_at=schedule)

    def test_jobschedule_maker(self):
        # basics
        sched = JobSchedule(minute='*')
        self.assertEqual(sched.text, 'Every minute')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, 'Every minute')
        # day specs
        sched = JobSchedule(weekday='mon-fri', at='06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        self.assertEqual(sched.cron, '00 06 * * mon-fri')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, sched.text)
        # day specs 2
        sched = JobSchedule(weekday='Mon-Fri', at='06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        self.assertEqual(sched.cron, '00 06 * * mon-fri')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, sched.text)
        # day specs 3
        sched = JobSchedule(weekday='Mon-FRI', at='06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        self.assertEqual(sched.cron, '00 06 * * mon-fri')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, sched.text)
        # day specs 4
        sched = JobSchedule(weekday='Mon-FRI', at='06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        self.assertEqual(sched.cron, '00 06 * * mon-fri')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, sched.text)
        # multiple times
        sched = JobSchedule(weekday='mon-fri', at='06:05,12:05')
        self.assertEqual(sched.text, 'At 06:05 AM and 12:05 PM, Monday through Friday')
        self.assertEqual(sched.cron, '05 06,12 * * mon-fri')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, sched.text)
        # months and days with every
        sched = JobSchedule(month='every 2', at='08:00', weekday='every 3')
        self.assertEqual(sched.text, 'At 08:00 AM, every 3 days of the week, every 2 months')
        self.assertEqual(sched.cron, '00 08 * */2 */3')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, sched.text)
        # step days
        sched = JobSchedule(weekday='every 2nd', at='06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, every 2 days of the week')
        self.assertEqual(sched.cron, '00 06 * * */2')
        sched = JobSchedule(weekday='every 1st', at='06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, only on Monday')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, sched.text)
        sched = JobSchedule(weekday='every 3rd', at='06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, every 3 days of the week')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, sched.text)
        sched = JobSchedule(weekday='every 4th', at='06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, every 4 days of the week')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, sched.text)
        # step hours
        sched = JobSchedule(hour='every 2nd', minute=0, weekday='mon-fri')
        self.assertEqual(sched.text, 'Every 2 hours, Monday through Friday')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, sched.text)
        # step minutes
        sched = JobSchedule(minute='every 5', weekday='mon-fri')
        self.assertEqual(sched.text, 'Every 5 minutes, Monday through Friday')
        sched2 = JobSchedule.from_cron(sched.cron)
        self.assertEqual(sched2.text, sched.text)
        # text specs
        sched = JobSchedule('friday, at 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, only on Friday')
        sched = JobSchedule('fridays, at 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, only on Friday')
        sched = JobSchedule('Mondays and Fridays, at 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, only on Monday and Friday')
        sched = JobSchedule('Mondays/Fridays, at 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, only on Monday and Friday')
        sched = JobSchedule('monday-friday, at 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        sched = JobSchedule('mon-fri, at 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        sched = JobSchedule('mon-fri at 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        sched = JobSchedule('Mon-Fri, at 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        sched = JobSchedule('Mon-Fri, 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        sched = JobSchedule('Mon-Fri 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        sched = JobSchedule('at 06:00, Mon-Fri')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        sched = JobSchedule('monday-friday, 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        sched = JobSchedule('every 2nd month, monday-friday, at 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday, every 2 months')
        sched = JobSchedule('every 5 minutes, every day, hour 6')
        self.assertEqual(sched.text, 'Every 5 minutes, at 06:00 AM')
        sched = JobSchedule('every 5 minutes every working day hour 6')
        self.assertEqual(sched.text, 'Every 5 minutes, at 06:00 AM, Monday through Friday')
        sched = JobSchedule('every 5 minutes every working day, hour 6')
        self.assertEqual(sched.text, 'Every 5 minutes, at 06:00 AM, Monday through Friday')
        sched = JobSchedule('every 5 minutes, every working day, hour 6')
        self.assertEqual(sched.text, 'Every 5 minutes, at 06:00 AM, Monday through Friday')
        sched = JobSchedule('every 5 minutes, on workdays, hours 6/7')
        self.assertEqual(sched.text, 'Every 5 minutes, at 06:00 AM and 07:00 AM, Monday through Friday')
        sched = JobSchedule('every 5 minutes, on workdays, in april')
        self.assertEqual(sched.text, 'Every 5 minutes, Monday through Friday, only in April')
        sched = JobSchedule('every 5 minutes, on weekends, in april')
        self.assertEqual(sched.text, 'Every 5 minutes, Saturday through Sunday, only in April')
        sched = JobSchedule('every 5 minutes, from monday to friday, in april')
        self.assertEqual(sched.text, 'Every 5 minutes, Monday through Friday, only in April')
        sched = JobSchedule('at 5 minutes, every hour, monday to friday, april')
        self.assertEqual(sched.text, 'At 5 minutes past the hour, Monday through Friday, only in April')
        sched = JobSchedule('1st day of month, at 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, on day 1 of the month')
        sched = JobSchedule('mon-fri, 06:00')
        self.assertEqual(sched.text, 'At 06:00 AM, Monday through Friday')
        sched = JobSchedule('every 2nd hour, 5 minute, weekdays')
        self.assertEqual(sched.text, 'At 5 minutes past the hour, every 2 hours, Monday through Friday')






import datetime
from notebook.services.contents.checkpoints import GenericCheckpointsMixin, Checkpoints


class NoOpCheckpoints(GenericCheckpointsMixin, Checkpoints):
    # source https://jupyter-notebook.readthedocs.io/en/stable/extending/contents.html#customizing-checkpoints
    # © Copyright 2015, Jupyter Team, https://jupyter.org. Revision 775cb20d.

    def checkpoint_model(self):
        info = dict(
            id='1',
            last_modified=datetime.datetime.now(),
        )
        return info

    def create_file_checkpoint(self, content, format, path):
        """ -> checkpoint model"""
        return self.checkpoint_model()

    def create_notebook_checkpoint(self, nb, path):
        """ -> checkpoint model"""
        return self.checkpoint_model()

    def get_file_checkpoint(self, checkpoint_id, path):
        """ -> {'type': 'file', 'content': <str>, 'format': {'text', 'base64'}}"""
        return {
            'type': 'file',
            'content': None,
            'format': format,
        }

    def get_notebook_checkpoint(self, checkpoint_id, path):
        """ -> {'type': 'notebook', 'content': <output of nbformat.read>}"""
        return {
            'type': 'notebook',
            'content': None,
        }

    def delete_checkpoint(self, checkpoint_id, path):
        """deletes a checkpoint for a file"""

    def list_checkpoints(self, path):
        """returns a list of checkpoint models for a given file,
        default just does one per file
        """
        return []

    def rename_checkpoint(self, checkpoint_id, old_path, new_path):
        """renames checkpoint from old path to new path"""





class BaseModelBackend(object):

    """
    OmegaML BaseModelBackend to be subclassed by other arbitrary backends

    This provides the abstract interface for any model backend to be implemented
    """
    def __init__(self, model_store=None, data_store=None, **kwargs):
        assert model_store, "Need a model store"
        assert data_store, "Need a data store"
        self.model_store = model_store
        self.data_store = data_store

    @classmethod
    def supports(self, obj, name, **kwargs):
        """
        test if this backend supports this obj
        """
        return False

    def get(self, name, **kwargs):
        """
        retrieve a model

        :param name: the name of the object
        :param version: the version of the object (not supported)
        """
        # support new backend architecture while keeping back compatibility
        return self.get_model(name, **kwargs)

    def put(self, obj, name, **kwargs):
        """
        store a model

        :param obj: the model object to be stored
        :param name: the name of the object
        :param attributes: attributes for meta data
        """
        # support new backend architecture while keeping back compatibility
        return self.put_model(obj, name, **kwargs)

    def put_model(self, obj, name, attributes=None, **kwargs):
        """
        store a model

        :param obj: the model object to be stored
        :param name: the name of the object
        :param attributes: attributes for meta data
        """
        raise NotImplementedError

    def get_model(self, name, version=-1, **kwargs):
        """
        retrieve a model

        :param name: the name of the object
        :param version: the version of the object (not supported)
        """
        raise NotImplementedError

    def predict(
            self, modelname, Xname, rName=None, pure_python=True, **kwargs):
        """
        predict using data stored in Xname

        :param modelname: the name of the model object
        :param Xname: the name of the X data set
        :param rName: the name of the result data object or None
        :param pure_python: if True return a python object. If False return
            a dataframe. Defaults to True to support any client.
        :param kwargs: kwargs passed to the model's predict method
        :return: return the predicted outcome   
        """
        raise NotImplementedError

    def predict_proba(
            self, modelname, Xname, rName=None, pure_python=True, **kwargs):
        """
        predict the probability using data stored in Xname

        :param modelname: the name of the model object
        :param Xname: the name of the X data set
        :param rName: the name of the result data object or None
        :param pure_python: if True return a python object. If False return
            a dataframe. Defaults to True to support any client.
        :param kwargs: kwargs passed to the model's predict method
        :return: return the predicted outcome   
        """
        raise NotImplementedError

    def fit(self, modelname, Xname, Yname=None, pure_python=True, **kwargs):
        """
        fit the model with data 

        :param modelname: the name of the model object
        :param Xname: the name of the X data set
        :param Yname: the name of the Y data set
        :param pure_python: if True return a python object. If False return
           a dataframe. Defaults to True to support any client.
        :param kwargs: kwargs passed to the model's predict method
        :return: return the meta data object of the model   
        """
        raise NotImplementedError

    def partial_fit(
            self, modelname, Xname, Yname=None, pure_python=True, **kwargs):
        """
        partially fit the model with data (online) 

        :param modelname: the name of the model object
        :param Xname: the name of the X data set
        :param Yname: the name of the Y data set
        :param pure_python: if True return a python object. If False return
           a dataframe. Defaults to True to support any client.
        :param kwargs: kwargs passed to the model's predict method
        :return: return the meta data object of the model   
        """

        raise NotImplementedError

    def fit_transform(
            self, modelname, Xname, Yname=None, rName=None, pure_python=True,
            **kwargs):
        """
        fit and transform using data

        :param modelname: the name of the model object
        :param Xname: the name of the X data set
        :param Yname: the name of the Y data set
        :param rName: the name of the transforms's result data object or None
        :param pure_python: if True return a python object. If False return
           a dataframe. Defaults to True to support any client.
        :param kwargs: kwargs passed to the model's transform method
        :return: return the meta data object of the model   
        """
        raise NotImplementedError

    def transform(self, modelname, Xname, rName=None, **kwargs):
        """
        transform using data

        :param modelname: the name of the model object
        :param Xname: the name of the X data set
        :param rName: the name of the transforms's result data object or None
        :param kwargs: kwargs passed to the model's transform method
        :return: return the transform data of the model   
        """
        raise NotImplementedError

    def score(
            self, modelname, Xname, Yname, rName=True, pure_python=True,
            **kwargs):
        """
        score using data

        :param modelname: the name of the model object
        :param Xname: the name of the X data set
        :param Yname: the name of the Y data set
        :param rName: the name of the transforms's result data object or None
        :param pure_python: if True return a python object. If False return
           a dataframe. Defaults to True to support any client.
        :param kwargs: kwargs passed to the model's predict method
        :return: return the score result   
        """
        raise NotImplementedError



from mongoengine import GridFSProxy

from omegaml import load_class
from omegaml.backends.basedata import BaseDataBackend


class ProtobufDataBackend(BaseDataBackend):
    KIND = 'protobuf.pbf'

    @classmethod
    def supports(self, obj, name, **kwargs):
        from google.protobuf.message import Message
        return isinstance(obj, Message)

    def put(self, obj, name, attributes=None, **kwargs):
        data = obj.SerializeToString()
        fileid = self.model_store.fs.put(data, filename=self.model_store._get_obj_store_key(name, '.pbf'))
        gridfile = GridFSProxy(grid_id=fileid,
                               db_alias='omega',
                               collection_name=self.model_store.bucket)
        kind_meta = {
            'protobuf_type': '{module}.{name}'.format(module=obj.DESCRIPTOR._concrete_class.__module__,
                                                      name=obj.DESCRIPTOR._concrete_class.__name__)
        }
        return self.model_store._make_metadata(
            name=name,
            prefix=self.model_store.prefix,
            bucket=self.model_store.bucket,
            kind=self.KIND,
            kind_meta=kind_meta,
            attributes=attributes,
            gridfile=gridfile).save()

    def get(self, name, version=-1, force_python=False, lazy=False, **kwargs):
        meta = self.data_store.metadata(name)
        data = meta.gridfile.read()
        # we get back a raw message, convert back to actual protobuf type
        # note the type must be loadable. will not work for arbitrary types
        pbtype = meta.kind_meta['protobuf_type']
        message = load_class(pbtype)()
        message.ParseFromString(data)
        return message

